"""
File Purpose: misc. tools related to checking xarray size,
and doing manipulations more efficiently with chunking for very large arrays.
"""
import numpy as np
import xarray as xr

from .xarray_accessors import pcAccessor
from .xarray_agg_stats import xarray_sum
from ..docs_tools import format_docstring
from ..math import product
from ..oop_tools import attach_attrs
from ..sentinels import UNSET
from ...defaults import DEFAULTS
from ...errors import (
    DimensionValueError,
    InputError,
    MemorySizeError,
)


### --------------------- Predict Size --------------------- ###

@pcAccessor.register('max_dim_sizes', totype='array')
def xarray_max_dim_sizes(array, *more_arrays, assert_broadcastable=False):
    '''return dict of {dim: max size across all arrays} for all dims in any of the arrays.

    assert_broadcastable: bool
        whether to raise DimensionValueError if any dim has different sizes in different arrays
        (other than size=1, which is considered broadcastable to any other size.)
        [TODO] more lenient definition which checks shared coordinates instead.
    '''
    arrays = (array,) + more_arrays
    result = {}
    for arr in arrays:
        for d, s in arr.sizes.items():
            if d in result:
                if result[d] != s:
                    if s == 1 or result[d] == 1:
                        result[d] = max(result[d], s)
                    elif assert_broadcastable:
                        errmsg = (f'cannot get max_dim_sizes due to different sizes for dim {d!r}: '
                                  f'{result[d]} vs {s}.')
                        raise DimensionValueError(errmsg)
                    else:
                        result[d] = max(result[d], s)
            else:
                result[d] = s
    return result

@pcAccessor.register('predict_result_size', totype='array')
def xarray_predict_result_size(array, *more_arrays, units='items'):
    '''return prediction of how large the product of these xarray objects would be.
    (or, any other math operation which broadcasts operations in the standard way..)

    units: str
        tells the units of the result (default: 'items'). Options:
        'items' --> number of items (elements) in the joined array.
        'bytes', 'kB', 'MB', 'GB' --> memory size of the data in joined array.
            (based on each array's dtype.itemsize and number of items.)
    '''
    arrays = (array,) + more_arrays
    VALID_UNITS = ('items', 'bytes', 'kB', 'MB', 'GB')
    if units not in VALID_UNITS:
        raise InputError(f'invalid units={units!r}. Expected one of {VALID_UNITS}.')
    result_dim_sizes = xarray_max_dim_sizes(*arrays, assert_broadcastable=True)
    n_items = product(result_dim_sizes.values())
    if units == 'items':
        return n_items
    # else, need to convert to bytes
    result_dtype = np.result_type(*arrays)
    n_bytes = n_items * result_dtype.itemsize
    bytes_per_unit = {'bytes': 1, 'kB': 1024, 'MB': 1024**2, 'GB': 1024**3}
    return n_bytes / bytes_per_unit[units]

@pcAccessor.register('result_size_check', totype='array')
def xarray_result_size_check(array, *more_arrays, GBmax=UNSET, safety=1, errmsg=None):
    '''ensure that the result of combining these xarrays (e.g., via product) is not too large.
    Too large if safety * predicted_size_GB > GBmax

    GBmax: UNSET, None, or number
        maximum size allowed, in Gigabytes.
        UNSET --> use DEFAULTS.RESULT_ARRAY_GBYTES_MAX
        None --> no limit (equivalent to "skip this check")
    safety: number
        safety factor; safety * predicted_size_GB is the actual limit.
        smaller is "safer" (effectively reduces limit)
    errmsg: None or str
        error message to use if the size is too large.
        will be formatted with .format(nGB=nGB, GBmax=GBmax, safety=safety,
                         dimsizes=(dict of predicted sizes of result dims)).

    raises MemorySizeError if array is too large.
    '''
    GBmax_is_default = (GBmax is UNSET)
    if GBmax is UNSET:
        GBmax = DEFAULTS.RESULT_ARRAY_GBYTES_MAX
    if GBmax is not None:
        nGB = xarray_predict_result_size(array, *more_arrays, units='GB')
        if safety * nGB > GBmax:
            dimsizes = xarray_max_dim_sizes(array, *more_arrays)
            if errmsg is None:
                errmsg = 'predicted size of result is too large: {nGB:.2f} GB > GBmax (={GBmax})'
                if safety != 1:
                    errmsg += ' / safety (={safety})'
                errmsg += '\nPredicted result dimsizes: {dimsizes}'
                if GBmax_is_default:
                    errmsg += '\nTo change limit, adjust DEFAULTS.RESULT_ARRAY_GBYTES_MAX.'
                errmsg += '\nTo skip this check, set the limit to None instead of a number.'
            errmsg = errmsg.format(nGB=nGB, GBmax=GBmax, safety=safety, dimsizes=dimsizes)
            raise MemorySizeError(errmsg)


### --------------------- Multiply Then Sum Along --------------------- ###

_paramdocs_xarray_multiply_then_sum_along = {
    'loop': '''None, str, or (str, int, mode).
        Indicates looping instructions (if any) to use during xarray_multiply_then_sum_along.
        None --> no looping; simply multiply all arrays, then sum along dim.
            Conceptually simplest. But, internally builds unnecessarily large array,
            which might be larger than DEFAULTS.RESULT_ARRAY_GBYTES_MAX and thus trigger MemorySizeError.
        str --> loop over each point along the indicated dimension.
            If loop != dim, sum at each point and then concatenate final result.
            If loop == dim, keep a running sum during the loop; no need to concatenate final result.
        'int' or 'sum' --> equivalent to using loop == dim.
        (str, integer, mode) --> (loop dimension, chunk size, chunk size mode)
            loop over the indicated dimension in chunks of the indicated size,
            instead of only one point at a time. This is usually the most efficient option. Mode can be:
                'GB' --> size of chunk arrays in GB (approximately)
                    E.g., ('logT', 5, 'GB') picks chunk size for logT such that each chunk is 5 GB or less.
                'size' --> number of points in the dimension.
                    E.g., ('x', 5, 'size') does first 5 points along x, then next 5, etc.
                'n' --> number of chunks (roughly)
                    E.g., ('x', 5, 'n') does first 1/5 of n points, then next 1/5, etc.
                    Not guaranteed to be exactly 5 chunks, due to rounding.''',
    'errmsg': '''None or str
        error message to use if the predicted size of any intermediate array is too big.
        Will be hit by .format(nGB=nGB, GBmax=GBmax, dimsizes=(dict of predicted sizes of result dims)).
        None --> make reasonably informative default error message.''',
}

_xarray_multiply_then_sum_along_errmsg = \
    '''In xarray_multiply_then_sum_along, multiplying arrays (before summing) would be too large!
    The limit is set by DEFAULTS.RESULT_ARRAY_GBYTES_MAX = {GBmax} GB.
    Predicted size: {nGB:.2f} GB, corresponding to dims with sizes (number of items):
        {dimsizes}.

    See help(xarray_multiply_then_sum_along) or help(array.pc.multiply_then_sum_along) for details.

    Options to fix this problem include any combination of the following:
        - work with smaller arrays. E.g., use slices or chunks appropriately.
        - adjust `loop` kwarg to loop over an internal dimension.
            See help(xarray_multiply_then_sum_along) for details.
        - (simple but not reommended) increase DEFAULTS.RESULT_ARRAY_GBYTES_MAX,
            or, set it to None to disable this check entirely.
            (Warning: might crash machine if internal array is too big!!)'''

@pcAccessor.register('multiply_then_sum_along', totype='array')
@format_docstring(**_paramdocs_xarray_multiply_then_sum_along)
@attach_attrs(_paramdocs_dict=_paramdocs_xarray_multiply_then_sum_along,
              _errmsg_default=_xarray_multiply_then_sum_along_errmsg)
def xarray_multiply_then_sum_along(array, *more_arrays, dim, loop=('int', 0.45, 'GB'), errmsg=None):
    '''returns result of product of arrays then .sum(dim=dim).
    Numerically equivalent to (arr0 * arr1 * ...).sum(dim=dim).

    But, internally does extra checks, and can be more efficient for large arrays:
        - ensures intermediate steps and final result never lead to "way too large" array,
            i.e. never bigger than DEFAULTS.RESULT_ARRAY_GBYTES_MAX.
            If it would be too big, crash with helpful MemorySizeError, instead.
        - may loop over a dimension, depending on `loop`, to avoid creating large intermediate arrays.

    array, more_arrays: xarray.DataArray objects
        the arrays to be multiplied
    dim: str
        the dimension to sum along after multiplying.
        (ignored if not present in any of the arrays.)
    loop: {loop}
    errmsg: {errmsg}
    '''
    # setup
    if errmsg is None:
        errmsg = xarray_multiply_then_sum_along._errmsg_default
    arrays = (array, *more_arrays)
    if loop is None:  # simplest, but also most likely to be too big
        xarray_result_size_check(*arrays, errmsg=errmsg)
        result = xarray_sum(product(arrays), dim=dim, missing_dims='ignore')
        return result
    else:
        ## bookkeeping ##
        if isinstance(loop, str):
            loop = (loop, 1, 'size')
        loopdimstr, loopsize, mode = loop
        if loopdimstr in ('int', 'sum'):
            x = dim   # x == loop dim.
        else:
            x = loopdimstr
        _valid_modes = ('size', 'n', 'GB')
        if mode not in _valid_modes:
            raise InputError(f"invalid loop mode: {mode!r}; expected one of {_valid_modes}")
        # size
        dim_sizes = xarray_max_dim_sizes(*arrays)
        if mode == 'size':
            size = loopsize    # size == size (number of points)
        elif mode == 'n':
            size = dim_sizes[x] // loopsize  # e.g. dimsize=110, n=25 --> size=4.
        elif mode == 'GB':
            full_GB = xarray_predict_result_size(*arrays, units='GB')
            # e.g. loopsize = 0.5 GB, full_GB = 2.3 --> size = int(dimsize / 5)
            size = int(dim_sizes[x] * min(1, loopsize / full_GB))  # min() in case loopsize > full_GB.
            if DEFAULTS.DEBUG >= 2:
                print(f'xarray_multiply_then_sum_along chose size={size} (of full={dim_sizes[x]}) '
                      f'for loopdim={x!r}, to make {full_GB * size/dim_sizes[x]:.3f} GB per chunk.')
        if size < 1:
            size = 1
        # slicers (based on size)
        if size == dim_sizes[x]:
            slicers = None  # no need to loop after all!
        elif size == 1:
            slicers = list(range(dim_sizes[x]))  # loop over each point
        else:
            slicers = []
            for i in range(0, dim_sizes[x], size):
                slicers.append(slice(i, min(i+size, dim_sizes[x]), None))
        # print final looping decisions:
        if DEFAULTS.DEBUG >= 3:
            print(f'xarray_multiply_then_sum_along using x={x!r}, size={size}, (due to mode={mode!r})')
        ## looping ##
        if slicers is None:  # not looping after all!
            xarray_result_size_check(*arrays, errmsg=errmsg)
            result = xarray_sum(product(arrays), dim=dim, missing_dims='ignore')
        elif x == dim:   # looping over the dimension being integrated along!
            for i, idx in enumerate(slicers):
                arrays_i = [a.isel({x: idx}, drop=True) if x in a.dims else a for a in arrays]
                if i == 0:
                    xarray_result_size_check(*arrays_i, errmsg=errmsg)
                    result = xarray_sum(product(arrays_i), dim=dim, missing_dims='ignore')
                else:
                    result += xarray_sum(product(arrays_i), dim=dim, missing_dims='ignore')
        else:  # looping over any other dimension (must concat at the end)
            result = []
            for i, idx in enumerate(slicers):
                arrays_i = [a.isel({x: idx}, drop=True) if x in a.dims else a for a in arrays]
                if i == 0:
                    xarray_result_size_check(*arrays_i, errmsg=errmsg)
                result.append(xarray_sum(product(arrays_i), dim=dim, missing_dims='ignore'))
            result = xr.concat(result, dim=x)  # concat along loop dimension
        return result



"""
File Purpose: logging, recording info about state of code, etc.

For reproducibility in the future.
E.g., with results, also save git hash for current version of the code.
"""

import datetime
import os
import subprocess
import sys

from .sentinels import UNSET


### --------------------- git_hash --------------------- ###

def git_hash_local(*, default=UNSET):
    '''returns the git commit hash for current git HEAD within the local directory

    default: any object
        if not UNSET, and can't get hash (but filepath does exist),
        return default & print message, instead of crashing if can't get hash.
    '''
    try:
        return subprocess.check_output(
            ['git', 'rev-parse', '--short', 'HEAD'],
            stderr=subprocess.DEVNULL,   # suppress crash message (crash gets handled below)
        ).decode('ascii').strip()
    except subprocess.CalledProcessError:
        if default is not UNSET:
            print(f'(failed to get git hash; returning default. At dir={repr(os.getcwd())})')
            return default
        raise

def git_hash(module_or_path='.', *, default=UNSET):
    '''returns the git commit hash for current git HEAD at the provided module or path.
    if string, treat it as a path; if module object, use path to module.__file__.
    module_or_path: string or module object, default '.'
        place to get git hash from.
        string --> path=module_or_path. If not a directory, also try os.path.dirname(path)
        module --> module=module_or_path. path=os.path.dirname(module.__file__)
    default: any object
        if provided, and can't get hash, return default & print message.
    '''
    try:
        if isinstance(module_or_path, str):
            path = module_or_path
        else:
            module = module_or_path
            path = module.__file__
        if not os.path.isdir(path):
            path = os.path.abspath(path)
            dirpath = os.path.dirname(path)
            if not os.path.isdir(dirpath):
                raise FileNotFoundError(f'{repr(path)} (not an existing directory). (\nalso {repr(dirpath)})')
            path = dirpath
        try:
            cwd0 = os.getcwd()
            os.chdir(path)
            return git_hash_local(default=default)
        finally:
            os.chdir(cwd0)
    except Exception:
        if default is not UNSET:
            print(f'(failed to get git hash; returning default. For input={repr(module_or_path)})')
            return default
        raise

def git_hash_here(globals_):
    '''returns git commit hash for the __file__ in the namespace where this function is called.'''
    return git_hash(globals_['__file__'])

def git_hash_PlasmaCalcs():
    '''returns the git commit hash for current git HEAD in PlasmaCalcs'''
    return git_hash_here(globals())


### --------------------- PlasmaCalcs version --------------------- ###

def _PlasmaCalcs_version():
    '''returns the current version of PlasmaCalcs, if available. Else, return None.'''
    try:
        from .. import __version__
    except ImportError:
        return None
    else:
        return __version__


### --------------------- datetime --------------------- ###

def datetime_astimezone(datetime_, tz='local'):
    '''convert a datetime.datetime to the requested timezone.
    Similar to datetime_.astimezone(tz), but also supports tz='local' and 'utc'.

    datetime_: datetime.datetime object
        datetime to convert.
    tz: 'local', 'utc', or datetime.tzinfo object
        timezone to convert result to.
        None --> do not convert.
        'local' --> machine's local timezone.
        'utc' --> standard/universal time (UTC) ("+00:00" if converted to str)
    '''
    if tz is None:
        pass
    elif tz == 'local':
        datetime_ = datetime_.astimezone()
    else:
        if tz == 'utc': tz = datetime.timezone.utc
        datetime_ = datetime_.astimezone(tz)
    return datetime_

def datetime_str(datetime_=None, *, sep="_", tz='local', timespec='seconds'):
    '''datetime as string, including timezone.
    Uses "isoformat", with "_" separator: YYYY-MM-DD_HH:MM:SS+HH:MM
    Datetime (timezone-aware) can be recovered via `datetime.datetime.fromisoformat(str)`

    datetime_: None or datetime.datetime object
        datetime to convert to string. If None, use datetime.datetime.now().
    sep: string
        separator between date and time. '_' nice when spaces not desireable (e.g. file names).
        'T' is nice for more universal recognizability; ' ' too but slightly less universal.
    tz: None, 'local', 'utc', or datetime.tzinfo object
        timezone to convert result to.
        None --> do not convert.
        'local' --> machine's local timezone.
        'utc' --> standard/universal time (UTC) (i.e. result will end in +00:00)
    timespec: str, probably 'seconds' or 'microseconds'
        passed to datetime.isoformat(). 'microseconds' --> YYYY-MM-DD_HH:MM:SS.ssssss+HH:MM
    '''
    if datetime_ is None:
        datetime_ = datetime.datetime.now()
    datetime_ = datetime_astimezone(datetime_, tz=tz)
    return datetime_.isoformat(sep=sep, timespec=timespec)

def datetime_now(*, sep="_", tz='local', timespec='microseconds'):
    '''current datetime as string according to user's computer. Include timezone.
    Uses "isoformat", with "_" separator: YYYY-MM-DD_HH:MM:SS.ssssss+HH:MM
    Datetime (timezone-aware) can be recovered via `datetime.datetime.fromisoformat(str)`

    sep: string
        separator between date and time. '_' nice when spaces not desireable (e.g. file names).
        'T' is nice for more universal recognizability; ' ' too but slightly less universal.
    tz: 'local', 'utc', or datetime.tzinfo object
        timezone to convert result to. 'local' --> machine's local timezone.
        'utc' --> standard/universal time (UTC) (i.e. result will end in +00:00)
    timespec: str, probably 'seconds' or 'microseconds'
        passed to datetime.isoformat(). 'microseconds' --> YYYY-MM-DD_HH:MM:SS.ssssss+HH:MM
    '''
    return datetime_str(datetime_=None, sep=sep, tz=tz, timespec=timespec)


### --------------------- file creation time (best effort) --------------------- ###

def datetimestamp(filepath, tz='local', *, as_str=False, sep="_"):
    '''return (best-effort guess of) datetime.datetime when this file was created or first written.
    Result depends on system:
        - macOS: real birth time
        - Windows: real creation time
        - Linux: falls back to modification time (st_mtime)

    filepath : str
        path to file
    tz: 'local', 'utc', or datetime.tzinfo object
        timezone to convert result to. 'local' --> machine's local timezone.
        'utc' --> standard/universal time (UTC) ("+00:00" if converted to str)
    as_str : bool
        whether to instead return as isoformat string using sep: YYYY-MM-DD_HH:MM:SS+HH:MM
    sep: string
        separator between date and time if as_str=True. (default '_' shown in format above^)
    '''
    stat = os.stat(filepath)

    # Platform-specific best-effort creation time
    if sys.platform == "win32":
        ts = stat.st_ctime  # creation time on Windows
    elif hasattr(stat, "st_birthtime"):
        ts = stat.st_birthtime  # macOS / BSD birth time
    else:
        # Linux fallback: use modification time
        ts = stat.st_mtime

    dt = datetime.datetime.fromtimestamp(ts, tz=datetime.timezone.utc)
    dt = datetime_astimezone(dt, tz=tz)

    if as_str:
        return datetime_str(dt, sep=sep, tz=None)  # tz=None because already converted above.
    else:
        return dt


### --------------------- PlasmaCalcs version --------------------- ###

def code_snapshot_info(*, default=None, skip_missing=False):
    '''returns dict including info about right now, and the version of the code being used.

    Includes as many of the following as possible (uses `default` if info unavailable due to crash):
        'pc__version__': PlasmaCalcs version
        'pc__commit_hash': git hash for PlasmaCalcs
        'datetime': current datetime, in isoformat

    default: any object, default None
        default value to use if info is unavailable due to crash.
    skip_missing: bool
        whether to skip missing info (i.e. any values which are None, probably due to crash).
    '''
    result = {}
    try:
        pc_version = _PlasmaCalcs_version()
    except Exception:
        pc_version = default
    if (not skip_missing) or (pc_version is not default):
        result['pc__version__'] = pc_version
    try:
        pc_hash = git_hash_PlasmaCalcs()
    except Exception:
        pc_hash = default
    if (not skip_missing) or (pc_hash is not default):
        result['pc__commit_hash'] = pc_hash
    try:
        datetime = datetime_now()
    except Exception:
        datetime = default
    if (not skip_missing) or (datetime is not default):
        result['datetime'] = datetime
    return result



"""
File Purpose: misc. tools related to registry
"""

import os
import sys
import uuid

from ..defaults import DEFAULTS
from ..errors import InputError
from ..tools import (
    UNSET,
    sql_location_string, sql_engine_from_location,
    _maybe_assert_sql_location_exists,
)


### --------------------- Registry database schema --------------------- ###

REGISTRY_SCHEMA_DIR = os.path.dirname(os.path.abspath(__file__))
REGISTRY_SCHEMA_FILES = {  # None has str value, all others have list values.
    None: os.path.join(REGISTRY_SCHEMA_DIR, 'registry_schema.sql'),
    'sqlite': [],  # <-- no extra files needed for sqlite.
    'postgres': [os.path.join(REGISTRY_SCHEMA_DIR, 'registry_schema_postgres.sql')],
}

def registry_schema_sql(engine=None, *, commit=True):
    '''load the SQL schema for the registry database, from registry_schema.sql.
    Returns a string containing the SQL commands that construct the database structure.
        (e.g. it will have CREATE TABLE commands and some CREATE INDEX commands).

    engine: None or str
        if provided, additionally load engine-specific commands.
        E.g. for 'postgres', combines registry_schema.sql with registry_schema_postgres.sql
    commit: bool
        whether to wrap the result as "BEGIN; ... ; COMMIT;"
    '''
    if engine not in REGISTRY_SCHEMA_FILES:
        raise InputError(f"unrecognized engine '{engine}' for registry schema")
    files = [REGISTRY_SCHEMA_FILES[None]]
    if engine is not None:
        files.extend(REGISTRY_SCHEMA_FILES[engine])
    result = []
    if commit:
        result.append('BEGIN;')
        result.append('\n')  # <-- just for style
    for filepath in files:
        result.append(f'\n\n/* -------- from {filepath!r} -------- */\n\n')
        with open(filepath, 'r', encoding='utf-8') as f:
            result.extend(f.readlines())
    if commit:
        result.append('\n')  # <-- just for style
        result.append('COMMIT;')
    return ''.join(result)


### --------------------- Registry location --------------------- ###

def registry_location_default(_platform=None):
    r'''get default registry location. This is the full path to the local pc_registry.db file,
    used by register() whenever DEFAULTS.REGISTRY_LOCATION=None and registry not provided.

    This is always that path, regardless of the current value of DEFAULTS.REGISTRY_LOCATION.
    
    The path is system-dependent. It will look something like this:
    - On Windows: C:\Users\<User>\AppData\Local\plasmacalcs\pc_registry.db
        (or, replace C:\Users\<User> with os.environ['LOCALAPPDATA'] if defined.)
    - On Mac: /Users/<User>/Library/Application Support/plasmacalcs/pc_registry.db
    - On Linux (or, any unrecognized platform): /Users/<User>/.plasmacalcs/pc_registry.db

    _platform: None or str
        if provided, return result computed as if sys.platform=_platform instead of its actual value.
        Mainly useful for debugging purposes.
        'win32' for windows, 'darwin' for Mac, anything else for Linux/Linux-like.
    '''
    if sys.platform == 'win32':
        base_dir = os.environ.get('LOCALAPPDATA',
            os.path.join(os.path.expanduser('~'), 'AppData', 'Local'))
        result = os.path.join(base_dir, 'plasmacalcs', 'pc_registry.db')
    elif sys.platform == 'darwin':
        base_dir = os.path.join(os.path.expanduser('~'), 'Library', 'Application Support')
        result =  os.path.join(base_dir, 'plasmacalcs', 'pc_registry.db')
    else:
        base_dir = os.path.expanduser('~')
        result = os.path.join(base_dir, '.plasmacalcs', 'pc_registry.db')
    return result

def registry_location(location=UNSET, *, assert_exists=None):
    '''get registry location implied by location input.
    Result is always an abspath or ":memory:".

    location: UNSET, None, or str
        UNSET --> use registry implied by DEFAULTS.REGISTRY_LOCATION
            (if DEFAULTS.REGISTRY_LOCATION=None, use registry_location_default(),
             which is system-dependent, e.g. "~/.plasmacalcs/pc_registry.db" on Linux.)
        None --> in-memory, equivalent to ":memory:". Will not edit any files.
        str --> Online database connection string starts with "postgresql://" or "postgres://".
                All other strings interpreted as filepath, and converted to abspath.
    assert_exists: None, True, or False
        True --> assert that implied location already exists as a file.
        False --> assert that implied location does not yet exist as a file.
        None --> no assertion is made.
        (Here, ':memory:' never "already exists", since it always means a new database.)
    '''
    if location is UNSET:
        result = DEFAULTS.REGISTRY_LOCATION
        if result is None:
            result = registry_location_default()
    else:
        result = sql_location_string(location)
    _maybe_assert_sql_location_exists(result, assert_exists=assert_exists)
    return result

def registry_name_default(location, name=UNSET, uuid_=None):
    '''get default registry name implied by name and location inputs.
    This is the default name that will be set, even if user didn't set one.
        (if name is provided as str, will return that str instead.)
    User can update registry name via 'UPDATE sources SET name = newname WHERE source_id = 0'.

    location: str
        ':memory:', abspath to database file, or online database connection string.
    name: UNSET, None, or str
        UNSET --> use basename of location. E.g. "pc_registry.db" by default.
            (if location in-memory, equivalent to name=None)
        None --> no name; use '_unnamed_<short_uuid>' as name,
            filling <short_uuid> with auto-generated unique hex string for this registry.
        str --> use this for registry name.
    uuid_: None or str
        if provided and need short_uuid, use uuid_[:8] instead of generating a new uuid.
    '''
    result = None
    if name is UNSET:
        engine = sql_engine_from_location(location)
        if engine == 'sqlite':
            result = os.path.basename(location)
        else:
            name = None  # get uuid name, below.
    if result is None:
        if name is None:
            if uuid_ is None: uuid_ = uuid.uuid4().hex
            result = f"_unnamed_{uuid_[:8]}"
        else:
            result = name
    return result



"""
File Purpose: misc. tools related to SQL.

Note that many of these solve the same issues that SQLAlchemy solves.
If you already know SQL well, maybe it would be better to use SQLAlchemy.
But, these tools were added so that PlasmaCalcs developers and users
can maybe learn SQL more easily, with less of the SQLAlchemy abstraction.
"""

import os
import contextlib
import logging
import sqlite3
from urllib.parse import urlparse

from .imports import ImportFailed

try:
    import psycopg
except ImportError as err:
    psycopg = ImportFailed("psycopg", err=err, locals=locals(), abbrv='psycopg')
try:
    import sqlglot
except ImportError as err:
    sqlglot = ImportFailed("sqlglot", err=err, locals=locals(), abbrv='sqlglot')

from .math import is_number 
from .sentinels import UNSET
from ..errors import InputError, InputConflictError


### --------------------- Misc helper methods --------------------- ###

def sql_location_string(location=None):
    '''returns "normalized" location string from location input.
    location: None or str
        None or ':memory:' --> in-memory, equivalent to ":memory:". Will not edit any files.
        str --> Online database connection string starts with "postgresql://" or "postgres://".
                All other strings interpreted as filepath, and converted to abspath.
    '''
    if location is None or location == ':memory:':
        return ':memory:'
    elif location.startswith('postgresql://') or location.startswith('postgres://'):
        return location
    else:
        return os.path.abspath(location)

def sql_engine_from_location(location):
    '''returns sql engine implied by location string.
    Possible results:
        "memory": sqlite3 in-memory connection
        "sqlite": sqlite3 connection to file
        "postgres": PostgreSQL connection to online database

    location: str
        implies sql connection, e.g. ":memory:", filepath, or online database connection string.
        postgresql connection string starts with "postgresql://" or "postgres://".

    Currently assumes any non-memory, non-postgres connection is a filepath,
        without checking whether the file exists yet.
    '''
    if location == ':memory:':
        return 'memory'
    elif location.startswith('postgresql://') or location.startswith('postgres://'):
        return 'postgres'
    else:
        return 'sqlite'

def sql_engine_backend(engine):
    '''return sql engine "backend" implied by engine string. Usually equivalent to engine string.
    But, some engines might have same backend, e.g., in-memory and file-based sqlite.

    The implementation here does:
        - 'memory' --> 'sqlite'
        - all other engines unchanged
    '''
    if engine == 'memory':
        return 'sqlite'
    else:
        return engine

def _maybe_assert_sql_location_exists(location, *, assert_exists=None):
    '''depending on assert_exists, may assert location exists or that it does not exist.

    location: str
        implies sql connection, e.g. ":memory:", filepath, or online database connection string.
        if filepath, should be input as abspath.
    assert_exists: None, True, or False
        True --> assert that implied location already exists as a file.
        False --> assert that implied location does not yet exist as a file.
        None --> no assertion is made.
        (Here, ':memory:' never "already exists", since it always means a new database.)
    '''
    if assert_exists is not None:
        if assert_exists is not True and assert_exists is not False:
            raise InputError(f"assert_exists must be None, True, or False, but got {assert_exists!r}")
        engine = sql_engine_from_location(location)
        if engine == 'memory':
            if assert_exists:
                raise InputConflictError("assert_exists=True but location==':memory:'.")
        elif engine == 'sqlite':
            exists = os.path.exists(location)
            if assert_exists is True and not exists:
                raise FileNotFoundError(f"assert_exists=True but database location does not exist: {location!r}")
            elif assert_exists is False and exists:
                raise FileExistsError(f"assert_exists=False but database location already exists: {location!r}")
        elif engine == 'postgres':
            try:
                with psycopg.connect(location):
                    pass
            except psycopg.OperationalError as err:
                if assert_exists is True:
                    errmsg = f"assert_exists=True but failed to connect to database at location: {location!r}."
                    raise FileNotFoundError(errmsg) from err
            else:
                if assert_exists is False:
                    errmsg = f"assert_exists=False but successfully connected to database at location: {location!r}."
                    raise FileExistsError(errmsg)

def sql_to_postgres(sql, *, dialect=UNSET, as_list=False, suppress_warnings=True):
    '''transpile sql statement into postgres.

    dialect: UNSET, None, or str
        The dialect corresponding to the input SQL.
        UNSET --> auto-detect; checks sqlite first, then postgres, then generic sql.
        None --> use generic sql, i.e. sqlglot's default dialect.
        str --> e.g., "sqlite", "postgres". Can be any dialect string recognized by sqlglot.
    as_list: bool
        whether to return result as a list of sql statements.
        False --> return as a single string, with statements separated by semicolons and newlines.
                The last statement will not have a semicolon at the end.
        True --> return a list of sql statements, none of which end with a semicolon.
    suppress_warnings: bool or int
        whether to suppress sqlglot warnings, e.g. about unsupported features of a dialect.
    '''
    try_dialects = ['sqlite', 'postgres', None] if dialect is UNSET else [dialect]
    with _sqlglot_suppress_warnings() if suppress_warnings else contextlib.nullcontext():
        for i, dial in enumerate(try_dialects):
            try:
                result = sqlglot.transpile(sql, read=dial, write='postgres')
            except Exception:
                if i == len(try_dialects) - 1:
                    raise  # last attempt, just raise the error if it fails.
            else:
                break
    return result if as_list else ';\n'.join(result)

@contextlib.contextmanager
def _sqlglot_suppress_warnings():
    '''context manager to suppress sqlglot warnings, e.g. about unsupported features of a dialect.'''
    logger = logging.getLogger("sqlglot")
    old_level = logger.level
    try:
        logger.setLevel(logging.ERROR)
        yield
    finally:
        logger.setLevel(old_level)


### --------------------- Connection Wrappers --------------------- ###

class Sqlite3InMemoryConnection():
    '''class which mimics a database connection to an in-memory database.
    Behaves like sqlite3.connect(':memory:') but will not close() upon __exit__.
    Allows consistent interface with "with" blocks without losing the in-memory connection,
        as long as the InMemoryConnection object itself is kept alive.

    con: None or sqlite3.Connection
        connection to in-memory database.
        If None, creates new sqlite3.Connection to ':memory:'.
    _owner_close_con_on_delete: bool
        whether self should try to close self.con when self is about to be deleted.
        Only applies if self created con, i.e. con not None during init.
    '''
    def __init__(self, *, con=None, _owner_close_con_on_delete=True):
        self._owns_con = con is None
        self._owner_close_con_on_delete = _owner_close_con_on_delete
        if con is None:
            con = sqlite3.connect(':memory:')
        con = sqlite3.connect(':memory:')
        if isinstance(con, Sqlite3InMemoryConnection):
            con = con.con
        self.con = con

    # # # CONTEXT MANAGER # # #
    def __enter__(self):
        '''enter "with" block; returns self (which behaves like sqlite3.Connection)'''
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        '''exit "with" block; does NOT close connection.'''
        pass

    def __del__(self):
        '''delete self; if self owns con and _owner_close_con_on_delete, close con.'''
        if self._owns_con and self._owner_close_con_on_delete:
            try:
                self.con.close()
            except Exception:  # it's fine if con was already closed...
                pass

    # # # OTHER CON-LIKE BEHAVIORS # # #
    def __getattr__(self, name):
        '''pass through any other attribute access to self.con'''
        return getattr(self.con, name)


class PostgresConnection():
    '''class which wraps a database connection to an online PostgreSQL database.
    Allows for sqlite3-like syntax, including executescript(), and using '?' placeholders.

    location: str
        online database connection string, starting with "postgresql://" or "postgres://".
    '''
    def __init__(self, location):
        self.location = location

    # # # CONTEXT MANAGER # # #
    def __enter__(self):
        '''enter "with" block; returns self (which behaves like psycopg connection)'''
        self.con = psycopg.connect(self.location)
        self.con.__enter__()
        return self

    def __exit__(self, exc_type, exc_value, traceback):
        '''exit "with" block; closes connection.'''
        self.con.__exit__(exc_type, exc_value, traceback)

    # # # TRANSLATION OF DIALECTS # # #
    sql_to_postgres = staticmethod(sql_to_postgres)

    # # # OTHER CON-LIKE BEHAVIORS # # #
    def __getattr__(self, name):
        '''pass through to self.con any attribute access not defined here.'''
        return getattr(self.con, name)

    def executescript(self, sql):
        '''execute multiple SQL statements at once, separated by semicolons.
        psycopg doesn't have executescript(), so we need to implement it ourselves.
        '''
        statements = sql_to_postgres(sql, as_list=True)
        for s in statements:
            self.con.execute(s)

    def execute(self, query, params=None, **kw_execute):
        '''execute SQL query. Can be written in sqlite-like syntax, e.g. using ? placeholders.'''
        pg_query = self.sql_to_postgres(query)
        return self.con.execute(pg_query, params, **kw_execute)


### --------------------- SqlBase --------------------- ###

class SqlBase():
    '''class which provides some basic SQL-related tools for a database connection.

    location: None or str
        None --> in-memory, equivalent to ":memory:". Will not edit any files.
        str --> connection string. Filepath for local sqlite connection.
                Online database connection string starts with "postgresql://" or "postgres://".
    assert_exists: None, True, or False
        True --> assert that implied location already exists as a file or online database.
        False --> assert that the implied location does not yet exist.
        None --> no assertion is made.
        (Here, ':memory:' never "already exists", since it always means a new database.)
    '''
    memory_connection_cls = Sqlite3InMemoryConnection
    postgres_connection_cls = PostgresConnection

    def __init__(self, location=None, *, assert_exists=None):
        self.location = sql_location_string(location)
        _maybe_assert_sql_location_exists(self.location, assert_exists=assert_exists)

    # # # SQL CONNECTION AND CROSS-ENGINE COMPATIBILITY # # #
    engine = property(lambda self: sql_engine_from_location(self.location),
            doc='''sql engine implied by location string. E.g. 'memory', 'sqlite', or 'postgres'.
            Might be more specific than just an sql dialect. See also: self.engine_backend.''')
    engine_backend = property(lambda self: sql_engine_backend(self.engine),
            doc='''sql engine "backend" implied by engine string. Corresponds to an sql dialect.
            E.g., 'sqlite' or 'postgres', but not 'memory' ('memory' engine <--> 'sqlite' backend).''')

    @property
    def is_local(self):
        '''returns whether self corresponds to a local database connection.
        Doesn't require the connection to actually exist. Currently just checks engine,
            e.g. 'postgres' --> is_local=False; 'memory' or 'sqlite' --> is_local=True.
        '''
        return self.engine in ['memory', 'sqlite']

    def _location_exists(self):
        '''return whether self.location already exists,
        as a cached in-memory connection, a file, or an online database.
        Caution: for online database, always returns True, i.e. assume it exists,
            because this code is not able to create a new online database from scratch.

        Subclass may want to implement self.schema_exists() which uses this method,
            and tells whether the relevant schema has been initialized already.
        '''
        engine = self.engine
        if engine == 'memory':
            return getattr(self, '_in_memory_con', None) is not None
        elif engine == 'sqlite':
            return os.path.exists(self.location)
        elif engine == 'postgres':
            return True

    def connect(self):
        '''return a connection to this database, creating it if it does not exist yet.
        (Cannot create online databases; those will just crash if connection fails.)

        Can be used as a context manager to ensure it is closed properly:
            with sqlbase.connect() as con:
                results = con.execute('SELECT ...').fetchall()
        Else, caller is responsible for closing the connection via con.close().
        '''
        engine = self.engine
        if engine == 'memory':
            # cache in-memory connection, so data persists for the lifetime of this Registry object.
            con = getattr(self, '_in_memory_con', None)
            exists = (con is not None)
            if not exists:
                con = self.memory_connection_cls()
                self._in_memory_con = con
        elif engine == 'sqlite':  # file-based database
            exists = os.path.exists(self.location)
            if not exists:  # makedirs to file if needed.
                os.makedirs(os.path.dirname(self.location), exist_ok=True)
            con = sqlite3.connect(self.location)
        elif engine == 'postgres':  # online database connection
            # assume we can connect, crash if we can't;
            # this code is not empowered to make a new online database.
            con = self.postgres_connection_cls(self.location)
        return con

    def execute(self, *args, **kwargs):
        '''shortcut for with self.connect() as con: return con.execute(*args, **kwargs)'''
        with self.connect() as con:
            return con.execute(*args, **kwargs)

    # # # SANITIZING # # #
    def sanitize_value(self, v):
        '''return sanitized value v, so that it can be stored in a SQL database.
        Sanitized means number, None, or string.
        '''
        if is_number(v):
            return v
        elif v is None:
            return None
        else:
            return str(v)

    def sanitize_values(self, d):
        '''return copy of dict d but with values sanitized to number, None, or string,
        so that they can be stored in a SQL database.
        '''
        sanitized = {}
        for k, v in d.items():
            sanitized[k] = self.sanitize_value(v)
        return sanitized

    # # # INSPECTION # # #
    def get_tables(self, *, _con=None):
        '''return list of all table names in this database.
        _con: None or open connection object
            for efficiency, can provide connection object if already created.
        '''
        with contextlib.ExitStack() as context_stack:
            if _con is None:
                con = self.connect()
                context_stack.enter_context(con)
            else:
                con = _con
            # sqlite3 solution (will need others for other engines):
            if self.engine_backend == 'sqlite':
                tables = con.execute("SELECT name FROM sqlite_master WHERE type='table'").fetchall()
                result = [row[0] for row in tables]
            else:
                assert self.engine_backend == 'postgres'
                tables = con.execute('''
                    SELECT table_name FROM information_schema.tables
                    WHERE table_schema = 'public'
                ''').fetchall()
                result = [row[0] for row in tables]
        return result

    @property
    def tables(self):
        '''list of all table names in this database.'''
        return self.get_tables()

    def columns(self, table):
        '''returns list of columns names in this table in this database.'''
        if self.engine_backend == 'sqlite':
            return [row[1] for row in self.execute(f'PRAGMA table_info({table})').fetchall()]
        else:
            assert self.engine_backend == 'postgres'
            columns = self.execute('''
                SELECT column_name FROM information_schema.columns
                WHERE table_name = %s
            ''', (table,)).fetchall()
            return [row[0] for row in columns]

    def column_types(self, table):
        '''returns the types of columns in this table in this database, as str-valued dict.'''
        if self.engine_backend == 'sqlite':
            return {row[1]: row[2] for row in self.execute(f'PRAGMA table_info({table})').fetchall()}
        else:
            assert self.engine_backend == 'postgres'
            columns = self.execute('''
                SELECT column_name, data_type FROM information_schema.columns
                WHERE table_name = %s
            ''', (table,)).fetchall()
            return {row[0]: row[1] for row in columns}

    def text_columns(self, table):
        '''returns the columns in this table whose types are text-based, e.g. TEXT, CHAR(32).'''
        column_types = self.column_types(table)
        text_cols = []
        for col, t in column_types.items():
            t = t.lower()
            for prefix in ['char', 'text', 'varchar']:
                if t.startswith(prefix):
                    text_cols.append(col)
                    break
        return text_cols

    def missing_columns(self, table, columns):
        '''returns list of which of these columns do not yet exist in table in this database.'''
        exists = set(self.columns(table))
        return [col for col in columns if col not in exists]

    def primary_keys(self, table):
        '''returns list of primary key column names in this table in this database.'''
        if self.engine_backend == 'sqlite':
            return [row[1] for row in self.execute(f'PRAGMA table_info({table})').fetchall() if row[5] == 1]
        else:
            assert self.engine_backend == 'postgres'
            pks = self.execute(f'''
                SELECT a.attname
                FROM   pg_index i
                JOIN   pg_attribute a ON a.attrelid = i.indrelid
                                    AND a.attnum = ANY(i.indkey)
                WHERE  i.indrelid = {table!r}::regclass
                AND    i.indisprimary;
            ''').fetchall()
            return [row[0] for row in pks]

    def primary_key(self, table):
        '''returns the single primary key column name in this table in this database,
        crashing if there is not exactly 1 primary key.
        '''
        result = self.primary_keys(table)
        if len(result) != 1:
            raise RegistryContentsError(f"Expected exactly 1 primary key in table {table!r}, but got: {result}")
        return result[0]

    @property
    def sizes(self):
        '''dict of sizes (number of entries) of all tables in this database.'''
        with self.connect() as con:
            tables = self.tables
            result = {}
            for table in tables:
                count = con.execute(f'SELECT COUNT(*) FROM {table}').fetchone()[0]
                result[table] = count
        return result

    # # # DISPLAY # # #
    def _location_display(self):
        '''return self.location but without any sensitive info, suitable for displaying in repr.
        For in-memory or filepath locations, just use self.location directly.
        For online database connection strings, just display engine and database name.
        '''
        if self.engine_backend == 'sqlite':
            return self.location
        else:
            assert self.engine_backend == 'postgres'
            parsed = urlparse(self.location)
            dbname = parsed.path.lstrip('/')
            hostname = parsed.hostname.split('.')[0] if parsed.hostname else 'unknownhost'
            return f'postgresql://{dbname}/host={hostname}'

    def _repr_contents(self):
        '''list of contents for self.__repr__'''
        return [f'location={self._location_display()!r}', f'sizes={self.sizes}']

    def __repr__(self):
        return f'{type(self).__name__}({", ".join(self._repr_contents())})'



"""
File Purpose MaindimsQuantitiesLoader
loading quantities related to the main dimensions
"""

from .quantity_loader import QuantityLoader
from ..dimensions import MainDimensionsHaver
from ..tools import xarray_np_gradient

class MaindimsQuantitiesLoader(MainDimensionsHaver, QuantityLoader):
    '''loader for quantities related to the main dimensions.'''

    @known_var
    def get_maindims_0(self):
        '''return array of 0s, with dimensions and coords implied by self.maindims.
        To save space, the result's dtype is int, by default.
        Equivalent: self.zeros_like_maindims()
        '''
        return self.zeros_like_maindims()

    @known_pattern(r'maindims_(.+)_coord')
    def get_maindims_dim_coord(self, var, *, _match=None):
        '''maindims_x_coord = array of x values and x coordinates.
        Equivalent: self.get_maindims_coord_xarrays()[x], self('maindims_0')[x]
        x must be one of self.maindims.

        Result's values use self.units; coords use self.coords_units (default==self.units).
            (e.g., units='cgs', coord_units='si' --> values are in cgs, but coords are in SI.)
        '''
        x, = _match.groups()
        self._assert_is_maindim(x)
        return self.get_maindims_coord_xarrays()[x]

    @known_pattern(r'maindims_(.+)_idx')
    def get_maindims_dim_idx(self, var, *, _match=None):
        '''maindims_x_idx == array of index associated with x values,
        based on self.slices. Only works if x in self.maindims.
        E.g. if slices[x]=slice(10, None, 4), then result values = [10, 14, 18, ...].
        Equivalent: self.get_maindims_coords_idx['x']
        '''
        x, = _match.groups()
        self._assert_is_maindim(x)
        return self.get_maindims_coords_idx()[x]

    @known_pattern(r'maindims_(.+)_didx', deps=[{0: 'maindims_{group0}_idx'}])
    def get_maindims_dim_didx(self, var, *, _match=None):
        '''maindims_x_didx == array of index spacing associated with x values,
        based on self.slices. Only works if x in self.maindims.
        E.g. if slices[x]=slice(10, None, 4), then result values = [4, 4, 4, ...].
        Equivalent: xarray_np_gradient(self('maindims_{x}_idx'), x)
        '''
        x, = _match.groups()
        idx = self(f'maindims_{x}_idx')  # asserts x is maindim
        return xarray_np_gradient(idx, x)

    @known_pattern(r'maindims_(.+)_dgrid')
    def get_maindims_dim_dgrid(self, var, *, _match=None):
        '''maindims_x_dgrid == array of underlying grid spacing associated with x values.
        Result's values use self.units; coords use self.coords_units (default==self.units).
            (e.g., units='cgs', coord_units='si' --> values are in cgs, but coords are in SI.)

        Caution: this is the underlying grid cell size, regardless of self.slices;
            Thus, this value is suitable for asking questions about the underlying grid,
                e.g. "what is the smallest wavelength the grid can support?"
            But it is NOT suitable for doing an integral across this dim, because, e.g.,
                setting self.slices[x]=slice(None,None,4) does NOT lead to 4 times larger result.
            For doing an integral, consider instead: self('maindims_{x}_dslice')

        Equivalent: self.get_maindims_dgrid(x)
        '''
        x, = _match.groups()
        self._assert_is_maindim(x)
        return self.get_maindims_dgrid(x)

    @known_pattern(r'maindims_(.+)_dslice',
                   deps=[{0: 'maindims_{group0}_dgrid'}, {0: 'maindims_{group0}_didx'}])
    def get_maindims_dim_dslice(self, var, *, _match=None):
        '''maindims_x_dslice == array of spacing between adjacent x values, accounting for self.slices.
        Result's values use self.units; coords use self.coords_units (default==self.units).
            (e.g., units='cgs', coord_units='si' --> values are in cgs, but coords are in SI.)

        Thus, this value is suitable for doing an integral.
            E.g., self.slices[x]=slice(None,None,4) should make ~4 times larger result than slice(None).
                (only "roughly", because the gradient accounts for adjacent values in the dgrid array.)
            For learning about the underlying grid, consider self('maindims_{x}_dgrid') instead.

        Equivalent: maindims_x_dgrid * maindims_x_didx.
        '''
        x, = _match.groups()
        dgrid = self(f'maindims_{x}_dgrid')
        didx = self(f'maindims_{x}_didx')
        return dgrid * didx



"""
File Purpose: "field-of-view" grid of rays through a box draped across a sphere.

Useful for making synthetic observations based on a plane-parallel (uncurved) simulation.
The overarching idea is:
    (1) Pick a FOV, like a grid of points in a rectangle on the sky.
    (2) Anchor the simulation box to the surface of a sphere.
    (3) "Drape" the box across the sphere, respecting curvature.
    (4) Trace rays from observer, through FOV, through draped box.

There are two "coordinate" systems (though, chosen coords can vary...) to keep in mind:
    (A) "SPH": describes the situation "as understood by the sphere".
        Rays are all straight lines in this system.
        Best coordinate systems for SPH:
            - spherical coordinates centered at sphere center
            - cartesian coordinates centered anywhere (but probably at sphere center)
    (B) "BOX": describes the situation "as understood by the (uncurved) simulation box".
        Rays are generally curved paths in this system.
        Best coordinate systems for BOX:
            - cartesian coordinates directly from simulation
            - cartesian coordinates centered at anchor point

There are various miscellaneous intracies:
    (1) Choosing FOV is simple enough. Using xarray makes it easy to have a grid of points.
    (2) Choosing anchor is simple enough, just need to pick a point in both coord systems.
        e.g. Abox=(x=X0+XEXT/2, y=Y0+YEXT/2, z=0); Asph=(r=R0, θ=(user choice), φ=(user choice)),
            where R0 = radius of sphere,
            XEXT, YEXT are x, y extents of the box,
            and X0, Y0 are x, y coordinates of lower-left corner of the box, default (0,0).
    (3) Draping the simulation is conceptually challenging but can be figured out,
        and the resulting formulas "aren't too bad". Here's what we want:
        (a) (going up in BOX) == (going same distance radially outward in SPH).
            --> zbox - Abox_z = rsph - Asph_r
            --> zbox = rsph - R0   (assuming Abox_z=0 and Asph_r=R0 as done in this file.)
        (b) (going horizontally in BOX at z=0) == (going same distance along sphere surface)
        Combining (a) and (b) almost fully specifies the mapping constraints,
        but there is one more choice to make: the direction for horizontal "motion".
        Basically, we need to pick two orthogonal unit vectors in the SPH system,
        in the plane tangent to the sphere at Asph ("tangent" ensures no changes in rsph).
        There are many ways to do this, but for this module let's choose:
        (c) (θhat_Asph <--> box x motion); (φhat_Asph <--> box y motion)
            Where unit vectors are defined in the usual way:
            rhat_Asph = sin(Asph_θ) cos(Asph_φ) xhat_sph
                       + sin(Asph_θ) sin(Asph_φ) yhat_sph
                       + cos(Asph_θ) zhat_sph
            θhat_Asph = cos(Asph_θ) cos(Asph_φ) xhat_sph
                       + cos(Asph_θ) sin(Asph_φ) yhat_sph
                       - sin(Asph_θ) zhat_sph
            φhat_Asph = - sin(Asph_φ) xhat_sph
                        + cos(Asph_φ) yhat_sph
        (TODO: allow any e1hat and e2hat for x and y, mixing θhat_Asph and φhat_Asph.)
        Now we have fully specified the mapping constraints.
    (4) For tracing rays, we need to know when the ray overlaps with the "draped box" region.
        Assuming we are allowed to periodically tile the box in x and y (and, around the sphere),
        this is easy enough. For a ray expressed parametrically as w(s) (vector),
        the ray overlaps the draped box region for all s such that:
            R0 <= |w(s)| <= R0 + ZEXT,
        where ZEXT is the extent of the simulation box above z=0.
        (If desired, the lower bound can be decreased to allow boxes which extend "below the surface".)
        This expression might be solved analytically to choose s,
        or it might just be imposed as a condition (e.g., mask values where it is False).

A bit more mapping logic:
    For clarity, let's first define Vsph:
            Vsph(xbox,ybox) = (xbox - Abox_x) θhat_Asph + (ybox - Abox_y) φhat_Asph
        Consider the tangent plane at Asph, before "draping" it onto the surface;
        the point in that plane corresponding to (xbox,ybox) in the simulation is:
            vsph(xbox,ybox,zbox=0) = Asph + Vsph(xbox,ybox)
        This is a conceptually useful definition which will reappear in the formulas.
    The mapping itself can now be determined algebraically by applying Rodrigues rotation formula,
        with angle:
            α = |Vsph(xbox,ybox)|/R0
        around axis of rotation:
            ghat_sph = rhat_Asph x Vsph_hat
            (where Vsph_hat = Vsph/|Vsph|)
        This determines the mapping because of (3b) above; "going a distance along the surface"
        corresponds to "rotating around by the correct amount, around the axis which is perpendicular
        to the plane containing the path as well as the sphere center." That axis is determined by
        rhat_Asph x Vhat_sph because those are two orthogonal vectors which lie in that plane.
    Finding the inverse mapping just involves algebra, starting from the mapping itself.
    
The mapping itself:
    ANCHOR POINT UNIT VECTORS (these are constants):
        rhat_Asph = sin(Asph_θ) cos(Asph_φ) xhat_sph
                   + sin(Asph_θ) sin(Asph_φ) yhat_sph
                   - cos(Asph_θ) zhat_sph
        θhat_Asph = cos(Asph_θ) cos(Asph_φ) xhat_sph
                   + cos(Asph_θ) sin(Asph_φ) yhat_sph
                   - sin(Asph_θ) zhat_sph
        φhat_Asph = - sin(Asph_φ) xhat_sph
                    + cos(Asph_φ) yhat_sph
    BOX TO SPH: (xbox,ybox,zbox) --> wsph (vector):
        rsph = R0 + zbox
        Vsph = (xbox - Abox_x) θhat_Asph + (ybox - Abox_y) φhat_Asph
        Vsph_hat = Vsph/|Vsph|
        α = |Vsph|/R0
        wsph = (cos(α) rhat_Asph + sin(α) Vsph_hat) rsph
    SPH TO BOX: wsph (vector) --> (xbox,ybox,zbox):
        rsph = |wsph|
        |Vsph| = R0 cos^-1((wsph . rhat_Asph)/rsph)
        wsph_perp = wsph - (wsph . rhat_Asph) rhat_Asph == rsph sin(α) Vsph_hat
        Vsph_hat = wsph_perp/|wsph_perp|
        Vsph = |Vsph| Vsph_hat
        xbox = Abox_x + (Vsph . θhat_Asph)
        ybox = Abox_y + (Vsph . φhat_Asph)
        zbox = rsph - R0
    TRICKY ANGLES:
        When wsph_perp --> 0, it is impossible to get Vsph_hat;
            this corresponds to the point xbox = Abox_x, ybox = Abox_y.
            --> we need a test like "if |wsph_perp| < ε, impose Vsph = 0.
                (maybe ε should be some small fraction of simulation grid cell size.)
        If (vsph . rhat_Asph)/rsph > 1, the inverse cosine will fail.
            this shouldn't occur anywhere though, |wsph| = |rsph| by definition,
            and rhat_Asph is a unit vector, so the whole thing should be 1 or less.
            Numerically might be an issue? Can double-check, but low priority.
        No other tricky angles, the rest should be fine!
    RAY FORMULA:
        Rays in sph system are straight lines:
            w(s) = F + rayhat * s
        where:
            rayhat = (F - T)/|F - T|,
            F is the point in FOV corresponding to this ray (there is one ray per FOV point),
            T is the telescope/observer position. (F and T both expressed in sph system.)
        The corresponding paths in box system are curved by the mapping formula above.
        Assuming periodic box tiling is allowed, the rays overlap the draped box region wherever:
            R0 <= |w(s)| <= R0 + ZEXT
        and the periodic box tiling can be accomplished by taking the modulus:
            xbox = xbox % XEXT
            ybox = ybox % YEXT
            zbox = zbox % ZEXT
    RAY BOUNDARIES:
        The values of s where rays enter and exit the draped box region can be computed from:
            w(s) = F + rayhat * s
            rayhat = (F - T)/|F - T|
            R0 <= |w(s)| <= R0 + ZEXT
        For boundary R = R0 or R0 + ZEXT:
            R^2 = F^2 + 2 (F . rayhat) s + s^2
            --> s^2 + 2 (F . rayhat) s + (F^2 - R^2) = 0
            --> s = - (F . rayhat) ± sqrt((F . rayhat)^2 - F^2 + R^2)
        The full ray intersects the boundary whenever the discriminant is non-negative.
        In that case, is important to know which point the ray passes through first,
            i.e. which is closest to T. That point is the "entry".
            If there are two intersections of a boundary, the other is the "exit".
            Thankfully, by definition, the smaller s value is always the entry point.
            (Not necessarily smaller in magnitude; s might be negative. E.g. -7 is smaller than 3.)
        The full decision tree looks like:
            - find lower and upper boundary entry and exit points. (Up to 4 distinct points).
            - if lower entry == lower exit, ray just grazes lower boundary; ignore lower boundary.
            - if upper entry == upper exit, ray just grazes upper boundary; replace ray with NaNs.
            - if no upper entry/exit, ray doesn't overlap sphere; replace ray with NaNs.
            - if no lower entry/exit, ray goes from upper entry to upper exit. (E.g., near disk edge.)
            - else, upper entry & exit, and lower entry & exit are all distinct,
                so ray goes from upper entry to lower entry (e.g., anywhere other than disk edge),
                and the exits both correspond to the draped region on the other side of the sphere.
        NOTE: when actually getting the standard ray_s values, the default is inclusive=(False, False),
            which avoids points "on the edges". This way all ray_s correspond to points INSIDE the box.
            Without doing this, some interp results may give NaNs due to points on edges.
    FOV POSITION AND SIZE:
        The FOV can be any size and be placed anywhere,
            as long as the sphere does not block it from T (not checked directly in this module...),
            and as long as it is a plane (not actually strictly necessary but probably a good idea).
        In practice, it is nice to choose FOV location based on box.
        How big should FOV be? Ideally, big enough to cover box.
        This defines "standard" FOV (assuming T along zsph axis):
            get points on box surfaces in sph system, project onto zsph=constant plane,
            then choose FOV x and y limits as min and max x and y across all projected points.
            Projecting all 8 corners is a good start but insufficient;
                also need to check top surface which can bulge out beyond the corners.
            Internal algorithm checks all 8 corners and a low-res grid (21x21 points) for top surface.
            (Currently, only implemented if T along zsph axis. Eventually, might handle arbitrary T.)
"""
import matplotlib.pyplot as plt
import numpy as np
import xarray as xr

from ..docs_tools import format_docstring
from ..math import is_integer
from ..properties import alias_child, simple_property
from ..xarray_tools import (
    xarray_sel,
    xr1d, xarray_grid,
    _get_basis, xarray_dot_product,
    xarray_magnitude, xarray_unit_vector,
    xarray_take_xyz, xarray_join_xyz,
    xarray_differentiate,
)
from ...errors import (
    InputError, InputMissingError, InputConflictError,
    TooManyPlottablesError,
    DimensionalityError,
)


### --------------------- Helper Classes (part of setup) --------------------- ###

class RayDrapeScene():
    '''representation of some "background" conditions for the ray draping problem.
    (High-level interface will allow user to change any of these,
        but for internal mental model, let's think of them as "probably constants".)
    Safest to convert to meters throughout, but it is probably technically units agnostic.

    R0: scalar
        radius of sphere (onto which z=0 of box will be draped)
    XEXT, YEXT, ZEXT: scalars
        extents of box to be draped
    T: scalar or xr.DataArray with 'component' dimension
        telescope/observer position (in sph system, cartesian coords centered on sphere)
        if scalar, interpret as distance along sph z axis, equivalent to T * pc.ZHAT.
        E.g., T = 1.5e11 * pc.ZHAT   # 1 AU from sphere center along sph z axis.
        (decent choice because it aligns zbox with zsph when viewing face-on / centered.)
    X0, Y0: scalar, default 0
        x, y coordinates of lower-left corner of box in box system
    '''
    _init_kwargs = {'R0', 'XEXT', 'YEXT', 'ZEXT', 'T', 'X0', 'Y0'}

    def __init__(self, *, R0, XEXT, YEXT, ZEXT, T, X0=0, Y0=0):
        if np.size(T) == 1:
            BASIS = _get_basis()
            ZHAT = BASIS.ZHAT
            T = T * ZHAT
        self.R0 = R0
        self.XEXT = XEXT
        self.YEXT = YEXT
        self.ZEXT = ZEXT
        self.T = T
        self.X0 = X0
        self.Y0 = Y0

    # # # DISPLAY # # #
    def _T_display(self):
        '''T in brief human-readable string if easy to do that, else just something reasonable.
        Extra pretty if 0.1 to 1000 AU (assuming SI units).
            (if non-SI units, might show AU accidentally, but this is just display,
            so it won't affect actual computations of results --> it's worth it.)
        '''
        T = self.T
        Tdisp = T.where(lambda arr: arr != 0, drop=True)
        if Tdisp.size == 1:
            comp = Tdisp['component'].values.item()
            zhat = f'{str(comp).upper()}HAT'
            return f"{Tdisp.values.item():.1e} {zhat}"
        elif T.size == 3:
            mag = np.sqrt(np.sum(T.values**2))
            return f"DataArray(vector with mag={mag:.1e})"
        else:
            return f"DataArray(sizes={dict(T.sizes)})"

    def _X0Y0_display(self):
        '''brief human-readable strings for X0 and Y0 excluding defaults (0).'''
        result = []
        for x in ['X0', 'Y0']:
            val = getattr(self, x)
            if val == 0:
                continue
            else:
                if np.size(val) > 1:
                    x_str = f"{type(val).__name__}(sizes={dict(val.sizes)})"
                else:
                    x_str = f"{val:.1e}"
                result.append(f"{x}={x_str}")
        return result

    def _repr_contents(self):
        '''list of contents for self.__repr__'''
        result = [
            f"R0={self.R0:.1e}",
            f"XEXT={self.XEXT:.1e}",
            f"YEXT={self.YEXT:.1e}",
            f"ZEXT={self.ZEXT:.1e}",
            f"T={self._T_display()}",
        ]
        result.extend(self._X0Y0_display())
        return result

    def __repr__(self):
        contents_str = ', '.join(self._repr_contents())
        return f'{type(self).__name__}({contents_str})'


class RayDrapeAnchor():
    '''representation of Anchor point for ray draping problem.
    For many helpful quantities, consider using self.Asph.
    (see help(type(self).Asph) for details.)

    Currently assumes Asph_r=R0, Abox_z=0.

    Asph_theta: scalar, default 0
        Asph_θ, i.e. anchor polar angle θ in sph system, in radians.
        This is the angle from +z to Asph, in the sph system.
        (for T along sph z axis, θ=0 corresponds to center of disk)
    Asph_phi: scalar, default 0
        Asph_φ, i.e. anchor azimuthal angle φ in sph system, in radians.
        This is the angle from +x to [Asph projected into x-y plane], in the sph system.
        (for T along sph z axis and θ=0, φ=0 provides intuitive x-y sph-box alignment,
            while φ=π/2 rotates box 90° clockwise as viewed from +z, thus "swapping" x and y)
    Abox_x: None or scalar
        Abox_x, i.e. anchor x position in box system
        None --> mapping will assume Abox_x = X0 + XEXT/2
    Abox_y: None or scalar
        Abox_y, i.e. anchor y position in box system
        None --> mapping will assume Abox_y = Y0 + YEXT/2
    '''
    # [TODO]: use UNSET for inputs with defaults; internally use class defaults for UNSET inputs.

    _init_kwargs = {'Asph_theta', 'Asph_phi', 'Abox_x', 'Abox_y'}

    def __init__(self, *, Asph_theta=0, Asph_phi=0, Abox_x=None, Abox_y=None):
        self.Asph_theta = Asph_theta
        self.Asph_phi = Asph_phi
        self.Abox_x = Abox_x
        self.Abox_y = Abox_y

    Asph_theta = simple_property('_Asph_theta',
        doc='''Asph_θ, i.e. anchor polar angle θ in sph system, in radians.
        See help(type(self)) for details.''')
    @Asph_theta.setter
    def Asph_theta(self, value):
        self._Asph_theta = value
        self._Astate = 1 + getattr(self, '_Astate', 0)
    
    Asph_phi = simple_property('_Asph_phi',
        doc='''Asph_φ, i.e. anchor azimuthal angle φ in sph system, in radians.
        See help(type(self)) for details.''')
    @Asph_phi.setter
    def Asph_phi(self, value):
        self._Asph_phi = value
        self._Astate = 1 + getattr(self, '_Astate', 0)

    @property
    def Asph(self):
        '''various details related to Asph_theta and Asph_phi.

        self.Asph is a dict containing:
            'θ': self.Asph_theta
            'cosθ': cos(Asph_theta)
            'sinθ': sin(Asph_theta)
            'φ': self.Asph_phi
            'cosφ': cos(Asph_phi)
            'sinφ': sin(Asph_phi)
            'rhat': rhat_Asph as xr.DataArray vector (along 'component' dim),
                    in sph system cartesian coords centered at r=0
            'θhat': θhat_Asph as xr.DataArray vector (along 'component' dim),
                    in sph system cartesian coords centered at r=0
            'φhat': φhat_Asph as xr.DataArray vector (along 'component' dim),
                    in sph system cartesian coords centered at r=0
        as well as alternate spellings of keys, with same values:
            'theta': alias to 'θ'
            'cos_theta': alias to 'cosθ'
            'sin_theta': alias to 'sinθ'
            'phi': alias to 'φ'
            'cos_phi': alias to 'cosφ'
            'sin_phi': alias to 'sinφ'
            'thetahat': alias to 'θhat'
            'phihat': alias to 'φhat'

        result is cached; only recomputed when Asph_theta or Asph_phi change.
        '''
        Astate, Asph = getattr(self, '_Asph_cached', (None, None))
        if Astate != self._Astate:
            Asph = self._Asph_calcs()
            self._Asph_cached = (self._Astate, Asph)
        return Asph

    def _Asph_calcs(self):
        '''compute various details related to Asph_theta and Asph_phi.
        See help(type(self).Asph) for details
        '''
        θ = self.Asph_theta
        φ = self.Asph_phi
        cosθ = np.cos(θ)
        sinθ = np.sin(θ)
        cosφ = np.cos(φ)
        sinφ = np.sin(φ)
        BASIS = _get_basis()
        XHAT = BASIS.XHAT
        YHAT = BASIS.YHAT
        ZHAT = BASIS.ZHAT
        rhat = (sinθ * cosφ) * XHAT + (sinθ * sinφ) * YHAT + (cosθ) * ZHAT
        θhat = (cosθ * cosφ) * XHAT + (cosθ * sinφ) * YHAT - (sinθ) * ZHAT
        φhat = (-sinφ) * XHAT + (cosφ) * YHAT
        result = {  # put all "nicer spellings" first, then all plain spellings.
            'θ': θ, 'cosθ': cosθ, 'sinθ': sinθ,
            'φ': φ, 'cosφ': cosφ, 'sinφ': sinφ,
            'rhat': rhat,
            'θhat': θhat, 'φhat': φhat,
            'theta': θ, 'cos_theta': cosθ, 'sin_theta': sinθ,
            'phi': φ, 'cos_phi': cosφ, 'sin_phi': sinφ,
            'thetahat': θhat, 'phihat': φhat,
        }
        return result

    def _get_basis(self):
        '''gets BasisXYZ object containing x,y,z basis info.
        Here, just returns _get_basis() which provides _BASIS from tools.xarray_tools.xarray_vectors
            (Component objects if imported the rest of PlasmaCalcs, strings if copy-pasted tools elsewhere.)
        '''
        return _get_basis()

    # # # DISPLAY # # #
    def _Asph_display(self):
        '''brief human-readable strings for Asph_θ and Asph_φ'''
        val = self.Asph_theta
        if np.size(val) > 1:
            θ_str = f"{type(val).__name__}(sizes={dict(val.sizes)})"
        else:
            θ_str = f"{np.degrees(val):.3g}°"
        val = self.Asph_phi
        if np.size(val) > 1:
            φ_str = f"{type(val).__name__}(sizes={dict(val.sizes)})"
        else:
            φ_str = f"{np.degrees(val):.3g}°"
        return [f"Asph_theta={θ_str}", f"Asph_phi={φ_str}"]

    def _Abox_display(self):
        '''brief human-readable strings for Abox_x and Abox_y, excluding defaults (None).'''
        result = []
        for Abox_x in ['Abox_x', 'Abox_y']:
            val = getattr(self, Abox_x)
            if val is None:
                continue
            else:
                if np.size(val) > 1:
                    x_str = f"{type(val).__name__}(sizes={dict(val.sizes)})"
                else:
                    x_str = f"{val:.1e}"
                result.append(f"{Abox_x}={x_str}")
        return result

    def _repr_contents(self):
        '''list of contents for self.__repr__'''
        result = self._Asph_display()
        result.extend(self._Abox_display())
        return result

    def __repr__(self):
        contents_str = ', '.join(self._repr_contents())
        return f'{type(self).__name__}({contents_str})'


class RayDrapeFOV():
    '''representation of Field-of-View grid for ray draping problem.

    FOV: xr.DataArray with 'component' dimension (and possibly any other dimensions too)
        grid of points representing "field of view" expressed in sph system,
        with 'component' dimension associated with x, y, z coords of each point.
    '''
    _init_kwargs = {'FOV'}

    def __init__(self, FOV):
        if not hasattr(FOV, 'dims') or 'component' not in FOV.dims:
            raise InputError("FOV must be an xarray object (e.g. DataArray) "
                             "with 'component' dimension associated with x, y, z coords of each point.")
        self.FOV = FOV

    # # # DISPLAY # # #
    def _repr_contents(self):
        '''list of contents for self.__repr__'''
        result = [f"FOV=DataArray{dict(self.FOV.sizes)}"]
        return result

    def __repr__(self):
        contents_str = ', '.join(self._repr_contents())
        return f'{type(self).__name__}({contents_str})'


### --------------------- RayDraping class --------------------- ###

_paramdocs_ray_draping = {
    'scene': '''dict or RayDrapeScene
        scene conditions for ray draping problem.
        dict --> convert to RayDrapeScene.
            Should have keys: R0, XEXT, YEXT, ZEXT, T.''',
    'anchor': '''dict or RayDrapeAnchor
        anchor point for ray draping problem.
        dict --> convert to RayDrapeAnchor.
            May have any of these keys: Asph_theta, Asph_phi, Abox_x, Abox_y.
            (Missing keys use RayDrapeAnchor defaults.)''',
    'fov': '''None, xr.DataArray, or RayDrapeFOV
        field-of-view grid for ray draping problem.
        None --> specify it later, by setting self.fov = value.
        xr.DataArray --> convert to RayDrapeFOV.
            Should have 'component' dimension associated with x, y, z coords of each point,
            in sph system cartesian coords centered at r=0.''',
    'wbox': '''xarray object (e.g. DataArray)
        points in box system cartesian coords,
        with 'component' dimension associated with x, y, z coords of each point.
        (Possibly useful: wbox=xr.concat([xbox, ybox, zbox], dim='component'))''',
    'wsph': '''xarray object (e.g. DataArray)
        points in sph system cartesian coords centered at r=0,
        with 'component' dimension associated with x, y, z coords of each point.
        (Possibly useful: wsph=xr.concat([xsph, ysph, zsph], dim='component'))''',
    'in_box': '''bool, 'lower', or 'upper'
        whether to check that points will fall inside the draped box region,
        masking all other points using NaNs.
        'lower' --> only check lower bound: R0 <= |wsph|.
        'upper' --> only check upper bound: |wsph| <= R0 + ZEXT.''',
    'periodic': '''bool
        whether to wrap points periodically into x and y box extents.
        True --> like draping the box periodically around the sphere.
        False --> mainly useful for debugging... are there other use cases?''',
}


@format_docstring(**_paramdocs_ray_draping)
class RayDrapingCore():
    '''core methods for the entire ray draping problem, including mapping calculations.
    This includes the setup (scene, anchor, fov), and mapping methods (for any points).

    Users should prefer RayDraping instead, which additionally includes
        the actual rays, some plotting methods, and some clever defaults.

    scene: {scene}
    anchor: {anchor}
    fov: {fov}
    '''
    scene_cls = RayDrapeScene
    anchor_cls = RayDrapeAnchor
    fov_cls = RayDrapeFOV

    # treat values as 0 if less than TINYTOL * np.finfo(...).eps,
    #  when doing the "tricky angles" check ("if |wsph_perp| < ε, impose Vsph = 0.")
    _TINYTOL = 1e2

    def __init__(self, *, scene, anchor, fov=None):
        if not isinstance(scene, RayDrapeScene):
            scene = self.scene_cls(**scene)
        if not isinstance(anchor, RayDrapeAnchor):
            anchor = self.anchor_cls(**anchor)
        self.scene = scene
        self.anchor = anchor
        self.fov = fov   # internally class self.fov_cls(fov) if needed; see type(self).fov.


    # # # MISC PROPERTIES # # #

    R0 = alias_child('scene', 'R0')
    XEXT = alias_child('scene', 'XEXT')
    YEXT = alias_child('scene', 'YEXT')
    ZEXT = alias_child('scene', 'ZEXT')
    T = alias_child('scene', 'T')
    X0 = alias_child('scene', 'X0')
    Y0 = alias_child('scene', 'Y0')
    Asph = alias_child('anchor', 'Asph')
    Abox_x = alias_child('anchor', 'Abox_x')
    Abox_y = alias_child('anchor', 'Abox_y')
    FOV = alias_child('fov', 'FOV')

    @Abox_x.getter
    def Abox_x(self):
        v = self.anchor.Abox_x
        return self.X0 + self.XEXT / 2 if v is None else v

    @Abox_y.getter
    def Abox_y(self):
        v = self.anchor.Abox_y
        return self.Y0 + self.YEXT / 2 if v is None else v

    _get_basis = alias_child('anchor', '_get_basis')


    # # # fov management # # #

    fov = simple_property('_fov',
        doc='''field-of-view grid for ray draping problem.
        For efficiency, RayDrapeFOV is created immediately when setting self.fov (unless None).
        (E.g., self.fov=arr; self.fov  # == RayDrapeFOV(arr))
        See help(type(self)) for more details about possible values.''')
    @fov.setter
    def fov(self, value):
        self._fov = self._get_fov_value(value)

    def _get_fov_value(self, value):
        '''convert value to RayDrapeFOV (unless None) appropriate for storage in self.fov.'''
        if value is None:
            return None
        elif not isinstance(value, RayDrapeFOV):
            value = self.fov_cls(value)
        return value

    @FOV.getter
    def FOV(self):
        '''alias to self.fov.FOV, this is the actual xr.DataArray associated with the FOV.
        Will raise InputMissingError if self.fov is None.
        '''
        if self.fov is None:
            raise InputMissingError("cannot get FOV when self.fov is None.")
        return self.fov.FOV


    # # # MAPPING - GENERIC FORMULAS # # #

    @format_docstring(**_paramdocs_ray_draping, sub_ntab=1)
    def box_to_sph(self, wbox):
        '''return wsph corresponding to any wbox point or points.
        i.e., returns w expressed in sph system cartesian coords centered at r=0.

        wbox: {wbox}
        '''
        xbox = xarray_sel(wbox, component='x')
        ybox = xarray_sel(wbox, component='y')
        zbox = xarray_sel(wbox, component='z')
        rsph = self.R0 + zbox
        Vsph = (xbox - self.Abox_x) * self.Asph['θhat'] + (ybox - self.Abox_y) * self.Asph['φhat']
        Vsph_mag = xarray_magnitude(Vsph)
        Vsph_hat = Vsph / Vsph_mag
        # check: "if |wsph_perp| < ε, impose Vsph = 0."
        TINY = np.finfo(Vsph_mag.dtype).eps * self._TINYTOL
        Vsph_hat = Vsph_hat.where(Vsph_mag > TINY, other=0)
        # (resume mapping math stuff)
        α = Vsph_mag / self.R0
        wsph = (np.cos(α) * self.Asph['rhat'] + np.sin(α) * Vsph_hat) * rsph
        return wsph

    @format_docstring(**_paramdocs_ray_draping, sub_ntab=1)
    def sph_to_box(self, wsph, *, in_box=False, periodic=True):
        '''return wbox corresponding to any wsph point or points.
        i.e., returns w expressed in box system cartesian coords.

        wsph: {wsph}
        in_box: {in_box}
            (Use False for efficiency if this was already checked elsewhere!)
        periodic: {periodic}
        '''
        # mapping math
        rsph = xarray_magnitude(wsph)
        wsph_dot_rhat = xarray_dot_product(wsph, self.Asph['rhat'])
        Vsph_mag = self.R0 * np.arccos(wsph_dot_rhat / rsph)
        wsph_perp = wsph - wsph_dot_rhat * self.Asph['rhat']
        wsph_perp_mag = xarray_magnitude(wsph_perp)
        Vsph_hat = wsph_perp / wsph_perp_mag
        Vsph = Vsph_mag * Vsph_hat
        # check: "if |wsph_perp| < ε, impose Vsph = 0."
        TINY = np.finfo(wsph_perp_mag.dtype).eps * self._TINYTOL
        Vsph = Vsph.where(wsph_perp_mag > TINY, other=0)
        # (resume mapping math stuff)
        xbox = self.Abox_x + xarray_dot_product(Vsph, self.Asph['θhat'])
        ybox = self.Abox_y + xarray_dot_product(Vsph, self.Asph['φhat'])
        zbox = rsph - self.R0
        # add component labels and join
        wbox = xarray_join_xyz(xbox, ybox, zbox, _basis=self._get_basis())
        # apply keep_in_box and periodic_wrap if desired
        if in_box:
            wbox = self.keep_in_box(wbox, in_box=in_box)
        if periodic:
            wbox = self.periodic_wrap(wbox)
        return wbox

    @format_docstring(**_paramdocs_ray_draping, sub_ntab=1)
    def keep_in_box(self, wsph, *, in_box=True):
        '''return wsph points masked to only those inside the draped box region.
        i.e., only keep points where R0 <= |wsph| <= R0 + ZEXT.

        wsph: {wsph}
        in_box: {in_box}
        '''
        rsph = xarray_magnitude(wsph)
        if in_box == 'lower':
            keep = (self.R0 <= rsph)
        elif in_box == 'upper':
            keep = (rsph <= self.R0 + self.ZEXT)
        elif in_box:
            keep = (self.R0 <= rsph) & (rsph <= self.R0 + self.ZEXT)
        wsph_kept = wsph.where(keep)
        return wsph_kept

    @format_docstring(**_paramdocs_ray_draping, sub_ntab=1)
    def periodic_wrap(self, wbox):
        '''return wbox points wrapped periodically into x and y box extents.
        The wrapping is done using modulus operation, accounting for X0 and Y0 if nonzero:
            xbox = (xbox - X0) % XEXT + X0
            ybox = (ybox - Y0) % YEXT + Y0

        wbox: {wbox}
        '''
        xbox = xarray_sel(wbox, component='x')
        ybox = xarray_sel(wbox, component='y')
        zbox = xarray_sel(wbox, component='z')
        if self.X0 == 0:  # <-- check not required, but can improve efficiency slightly.
            xbox = xbox % self.XEXT
        else:
            xbox = (xbox - self.X0) % self.XEXT + self.X0
        if self.Y0 == 0:  # <-- check not required, but can improve efficiency slightly.
            ybox = ybox % self.YEXT
        else:
            ybox = (ybox - self.Y0) % self.YEXT + self.Y0
        wbox_wrapped = xr.concat([xbox, ybox, zbox], dim='component')
        return wbox_wrapped


    # # # DISPLAY # # #

    def _repr_contents(self):
        '''list of contents for self.__repr__'''
        result = [
            f"scene={repr(self.scene)}",
            f"anchor={repr(self.anchor)}",
            f"fov={repr(self.fov)}",
        ]
        return result

    def __repr__(self):
        contents_str = ', '.join(self._repr_contents())
        return f'{type(self).__name__}({contents_str})'


_paramdocs_ray_draping.update({
    'ray_s': '''None, int, dict, or xr.DataArray including 'ray_s' or 'ray_s_dim' dimension.
        Distance along each ray, from FOV. Rays are parametrized (in sph system) as:
            w(s) = F + rayhat * s,   where:
            rayhat = (F - T)/|F - T|,
            F is the point in FOV corresponding to this ray (there is one ray per FOV point),
            T is the telescope/observer position.
        Must include the 'ray_s' or 'ray_s_dim' dimension. Can be 1D if all rays have same spacing,
            but might also have same dims as FOV, to specify different spacing for each ray.
        None --> specify later (getting self.ray_s before specifying will just cause crash).
        int --> number of points along each ray, choosing self.standard_ray_s_limits() for min and max s,
            while using inclusive=(False,False) when generating the grid (to avoid including edge points).
            standard_ray_s_limits() picks s limits to precisely cover full paths through the draped box region.
        dict --> kwargs for pc.xarray_grid(). Uses self.standard_ray_s_limits() for min and max, unless provided.
            Also uses name='ray_s' unless provided, and inclusive=(False,False) unless provided.
            [TODO] add "maxstep" option to xarray_grid(), to enable "maxstep = simulation dx".
            Workaround to solve that problem for now: specify ray_s=None when making drape, then later do:
                smin, smax = drape.standard_ray_s_limits()
                ray_s=pc.xarray_grid(smin, smax, N=1+int(np.max((smax-smin)/dx)))
        Examples: 100; dict(N=100); pc.xr1d([100, 200, 300, 400], name='ray_s').
        (Use self.ray_s to see ray_s value, and self.ray_s_array to see the actual ray_s xr.DataArray.)''',
})

@format_docstring(**_paramdocs_ray_draping)
class RayDraping(RayDrapingCore):
    '''representation of the entire ray draping problem, including mapping calculations.
    Consider using self.wbox(), to get "the answer", i.e. ray points mapped into box system.

    scene: {scene}
    anchor: {anchor}
    fov: 3-tuple of ("standard", Nx, Ny), {fov}
        ("standard", Nx, Ny) --> use self.standard_fov(Nx, Ny) to create reasonable FOV,
            chosen such as the rectangle at zsph=R0+ZEXT) which fully covers entire box
            (without any smart rotations yet; edges parallel to x sph and y sph axes).
    ray_s: {ray_s}


    EXAMPLE
    -------
        # Below is a somewhat exaggerated example for visualization purposes;
        #   if the 3D simulation box is really that large compared to the Sun,
        #   then curvature effects might be physically significant within the plasma,
        #   so maybe a plane-parallel simulation isn't the best idea on that scale.
        import PlasmaCalcs as pc
        import numpy as np
        Rsun = 6.957e8  # [m]
        AU = pc.DEFAULTS.PHYSICAL.CONSTANTS_SI['AU']   # 1.495978707e11  [m]
        scene = pc.RayDrapeScene(R0=Rsun,
                                 XEXT=1.0*Rsun,
                                 YEXT=0.7*Rsun,
                                 ZEXT=1.0*Rsun,
                                 T=1*AU)
        anchor = pc.RayDrapeAnchor(Asph_theta=np.pi*3/8,
                                   Asph_phi=np.pi/4)
        drape = pc.RayDraping(scene=scene, anchor=anchor,
                              fov=("standard", 7, 4),
                              ray_s=30)
        drape.plot('both', periodic=False)
    '''

    def __init__(self, *, scene, anchor, fov=None, ray_s=None):
        super().__init__(scene=scene, anchor=anchor, fov=fov)
        self.ray_s = ray_s


    @classmethod
    @format_docstring(scene_kw=RayDrapeScene._init_kwargs,
                      anchor_kw=RayDrapeAnchor._init_kwargs,
                      fov_kw=RayDrapeFOV._init_kwargs)
    def from_dict(cls, d):
        '''create RayDraping instance from dict, with keys corresponding to __init__ kwargs,
        or to kwargs used to initialize scene, anchor, and fov, e.g. can provide "T" here.

        Cannot overlap __init__ and scene/anchor/fov kwargs;
            e.g. cannot provide both "T" and "scene".

        Keys allowed are:
            'scene' or scene kwargs: {scene_kw}
            'anchor' or anchor kwargs: {anchor_kw}
            'fov' or fov kwargs: {fov_kw}
            'ray_s'
        '''
        scene_kw = cls.scene_cls._init_kwargs
        anchor_kw = cls.anchor_cls._init_kwargs
        fov_kw = cls.fov_cls._init_kwargs
        # checks:
        if 'scene' in d and any(k in d for k in scene_kw):
            overlap = tuple(k for k in scene_kw if k in d)
            raise InputConflictError(f"Cannot provide both 'scene' and scene kwargs {overlap}.")
        if 'anchor' in d and any(k in d for k in anchor_kw):
            overlap = tuple(k for k in anchor_kw if k in d)
            raise InputConflictError(f"Cannot provide both 'anchor' and anchor kwargs {overlap}.")
        if 'fov' in d and any(k in d for k in fov_kw):
            overlap = tuple(k for k in fov_kw if k in d)
            raise InputConflictError(f"Cannot provide both 'fov' and fov kwargs {overlap}.")
        # method:
        d = {**d}  # copy to avoid modifying input
        scene = d.pop('scene', None)
        if scene is None:
            scene = {k: d.pop(k) for k in scene_kw if k in d}
            scene = cls.scene_cls(**scene)
        anchor = d.pop('anchor', None)
        if anchor is None:
            anchor = {k: d.pop(k) for k in anchor_kw if k in d}
            anchor = cls.anchor_cls(**anchor)
        fov = d.pop('fov', None)
        if fov is None:
            fov = {k: d.pop(k) for k in fov_kw if k in d}
            if fov:
                fov = cls.fov_cls(**fov)
            else:
                fov = None  # if no fov kwargs provided, use None (i.e. specify later)
        ray_s = d.pop('ray_s', None)
        if len(d) > 0:
            raise InputError(f"Unrecognized keys in input dict: {list(d.keys())}")
        return cls(scene=scene, anchor=anchor, fov=fov, ray_s=ray_s)


    # # # MISC PROPERTIES # # #

    @property
    def n_rays(self):
        '''number of rays, i.e. number of points in self.FOV'''
        sizes = self.FOV.sizes
        n_rays = 1
        for dim in sizes:
            if dim != 'component' and dim != 'ray_s' and dim != 'ray_s_dim':
                n_rays = n_rays * sizes[dim]
        return n_rays


    # # # MAPPING - THESE PARTICULAR RAY_S # # #

    def rayhat(self):
        '''rayhat, i.e. unit vector along each ray in sph system cartesian coords.
        
        rayhat has dimensions of self.FOV,
            as well as 'component' dimension associated with x, y, z coords of each point.

        rayhat = (F - T)/|F - T|,   where:
            F is the point in FOV corresponding to this ray (there is one ray per FOV point),
            T is the telescope/observer position.
        '''
        F = self.FOV
        T = self.T
        return xarray_unit_vector(F - T)

    @format_docstring(**_paramdocs_ray_draping, sub_ntab=1)
    def wsph(self, *, in_box=True):
        '''return wsph, i.e. the points along each ray in sph system cartesian coords centered at r=0.
        
        wsph has dimensions of self.FOV plus 'ray_s' or 'ray_s_dim' dimension,
            as well as 'component' dimension associated with x, y, z coords of each point.

        wsph = F + rayhat * s,   where:
            rayhat = (F - T)/|F - T|,
            F is the point in FOV corresponding to this ray (there is one ray per FOV point),
            T is the telescope/observer position.

        in_box: {in_box}
        '''
        s = self.ray_s_array()
        F = self.FOV
        rayhat = self.rayhat()
        wsph = F + rayhat * s
        if in_box:
            wsph = self.keep_in_box(wsph, in_box=in_box)
        return wsph

    @format_docstring(**_paramdocs_ray_draping, sub_ntab=1)
    def wbox(self, *, in_box=True, periodic=True):
        '''return wbox, i.e. the points along each ray, in box system cartesian coords.
        These are the curved paths which are the "final answer" for the ray draping problem.

        wbox has dimensions of self.FOV plus 'ray_s' or 'ray_s_dim' dimension,
            as well as 'component' dimension associated with x, y, z coords of each point.

        in_box: {in_box}
        periodic: {periodic}
        '''
        wsph = self.wsph(in_box=in_box)
        wbox = self.sph_to_box(wsph, periodic=periodic,
                               in_box=False)  # [EFF] no need to re-check in_box here.
        return wbox

    @format_docstring(**_paramdocs_ray_draping, sub_ntab=1)
    def rays(self, system='box', *, in_box=True, periodic=True):
        '''return rays, i.e. points along each ray, in desired coordinate system.
        Result is an xr.DataArray with dims from self.FOV, 'ray_s' or 'ray_s_dim',
            and 'component' dim associated with x, y, z coords of each point.
    
        system: 'box' or 'sph'
            'box' --> return points in box system cartesian coords.
            'sph' --> return points in sph system cartesian coords centered at r=0.
        in_box: {in_box}
        periodic: {periodic}
            (only used if system=='box')

        Equivalent to self.wbox() or self.wsph(), depending on system.
        '''
        if system == 'box':
            rays = self.wbox(in_box=in_box, periodic=periodic)
        elif system == 'sph':
            rays = self.wsph(in_box=in_box)
        else:
            raise InputError(f"Expected system = 'box' or 'sph'; got {system!r}.")
        return rays

    def ray_tangents(self, system='box', *, unit_vector=False):
        '''return array of vectors pointing in ray directions (i.e. tangent to ray path).
        The implementation here takes derivatives along the 'ray_s' or 'ray_s_dim' dimension,
            which may produce artifacts at edges of ray_s.

        system: 'box' or 'sph'
            'box' --> return directions along rays in box system
            'sph' --> equivalent to self.rayhat() (which is a constant for each ray)
        unit_vector: bool
            whether to convert to unit vector.
            ('sph' result will always be unit vector, regardless.)
            False --> 'box' result will just be derivative of ray with respect to ray_s.
        '''
        if system == 'sph':
            return self.rayhat()
        else:
            rays = self.rays(system=system, periodic=False)
            return xarray_differentiate(rays, 'ray_s')


    # # # "STANDARD" FOV CHOICES # # #

    def _get_box_corners(self, system='box'):
        '''return all 8 corners of the box, as xr.DataArray with 'corner' and 'component' dims.
        Corners are in box system unless system=='sph', in which case converted via self.box_to_sph().
        '''
        BASIS = self._get_basis()
        XHAT = BASIS.XHAT
        YHAT = BASIS.YHAT
        ZHAT = BASIS.ZHAT
        corners = []
        for ix in [0, self.XEXT]:
            for iy in [0, self.YEXT]:
                for iz in [0, self.ZEXT]:
                    corner = ix * XHAT + iy * YHAT + iz * ZHAT
                    corners.append(corner)
        corners = xr.concat(corners, dim='corner')
        if system == 'sph':
            corners = self.box_to_sph(corners)
        return corners

    def _get_box_top(self, system='box', *, Nu=21, Nv=21):
        '''return top surface of box, with Nu x Nv points,
        along dims _pc_surface_u_ and _pc_surface_v_.
        Points are in box system unless system=='sph', in which case converted via self.box_to_sph().
        '''
        uu = xr1d(np.linspace(0, self.XEXT, Nu), name='_pc_surface_u_')
        vv = xr1d(np.linspace(0, self.YEXT, Nv), name='_pc_surface_v_')
        BASIS = self._get_basis()
        result = uu * BASIS.XHAT + vv * BASIS.YHAT + self.ZEXT * BASIS.ZHAT
        if system == 'sph':
            result = self.box_to_sph(result)
        return result

    def standard_fov_limits(self):
        '''return (xmin, xmax, ymin, ymax) for standard FOV.
        This is chosen from scene and anchor params,
            as rectangle (at zsph=R0+ZEXT) with size chosen to cover entire box,
            and edges parallel to x sph and y sph axes.
            It assumes that T is along sph z axis (else crash with NotImplementedError)

        See help(PlasmaCalcs.ray_draping) FOV POSITION AND SIZE for more details.
        See also: self.standard_FOV(), self.standard_fov()
        '''
        # check T along z sph axis.
        T = self.T
        T_hat = xarray_unit_vector(T)
        T_hatx = xarray_sel(T_hat, component='x')
        T_haty = xarray_sel(T_hat, component='y')
        if not np.allclose(T_hatx, 0) or not np.allclose(T_haty, 0):
            raise NotImplementedError("standard_fov_limits() only implemented for T along sph z axis.")
        # projecting to "perp to z" is easy, just ignore the z component.
        # corners
        corners = self._get_box_corners(system='sph')
        corners_x = xarray_sel(corners, component='x')
        x_min = corners_x.min().item()
        x_max = corners_x.max().item()
        corners_y = xarray_sel(corners, component='y')
        y_min = corners_y.min().item()
        y_max = corners_y.max().item()
        # top
        top = self._get_box_top(system='sph')
        top_x = xarray_sel(top, component='x')
        x_min = min(x_min, top_x.min().item())
        x_max = max(x_max, top_x.max().item())
        top_y = xarray_sel(top, component='y')
        y_min = min(y_min, top_y.min().item())
        y_max = max(y_max, top_y.max().item())
        return (x_min, x_max, y_min, y_max)

    def standard_FOV(self, Nx, Ny, *, xdim='FOVx', ydim='FOVy'):
        '''return standard FOV (array) with (Nx, Ny) points.
        This is chosen from scene and anchor params,
            as rectangle (at zsph=R0+ZEXT) with size chosen to cover entire box,
            and edges parallel to x sph and y sph axes.
            It assumes that T is along sph z axis (else crash with NotImplementedError)

        See help(PlasmaCalcs.ray_draping) FOV POSITION AND SIZE for more details.
        See also: self.standard_fov(), self.standard_fov_limits()
        '''
        xmin, xmax, ymin, ymax = self.standard_fov_limits()
        Fx = xr1d(np.linspace(xmin, xmax, Nx), name=xdim)
        Fy = xr1d(np.linspace(ymin, ymax, Ny), name=ydim)
        Fz = self.R0 + self.ZEXT
        BASIS = self._get_basis()
        FOV = Fx * BASIS.XHAT + Fy * BASIS.YHAT + Fz * BASIS.ZHAT
        return FOV

    def standard_fov(self, Nx, Ny, *, xdim='FOVx', ydim='FOVy'):
        '''return standard fov (RayDrapeFOV) with (Nx, Ny) points.
        This is chosen from scene and anchor params,
            as rectangle (at zsph=R0+ZEXT) with size chosen to cover entire box,
            and edges parallel to x sph and y sph axes.
            It assumes that T is along sph z axis (else crash with NotImplementedError)

        See help(PlasmaCalcs.ray_draping) FOV POSITION AND SIZE for more details.
        See also: self.standard_FOV(), self.standard_fov_limits()
        '''
        FOV = self.standard_FOV(Nx, Ny, xdim=xdim, ydim=ydim)
        return self.fov_cls(FOV)

    def _get_fov_value(self, value):
        '''convert value to RayDrapeFOV (unless None) appropriate for storage in self.fov.'''
        if isinstance(value, tuple) and len(value) == 3 and value[0] == 'standard':
            Nx = value[1]
            Ny = value[2]
            value = self.standard_fov(Nx, Ny)
        else:
            value = super()._get_fov_value(value)
        return value


    # # # "STANDARD" RAY CHOICES # # #

    def standard_ray_s_limits(self):
        '''return (min, max) s to use for self.ray_s, corresponding to full overlap with the box.
        Each ray may have its own (min, max) pair, so each is an xarray with same dims as self.FOV.

        See help(PlasmaCalcs.ray_draping) RAY BOUNDARIES section for more details.

        Value will be NaN for rays which never enter the draped box region.
        '''
        rayhat = self.rayhat()
        F = self.FOV
        F_dot_rayhat = xarray_dot_product(F, rayhat)
        F_dot_rayhat_2 = F_dot_rayhat**2
        F2 = xarray_magnitude(F)**2
        # upper boundary:
        r_upper = self.R0 + self.ZEXT
        D_upper = F_dot_rayhat_2 - F2 + r_upper**2  # discriminant
        upper_exists = (D_upper > 0)  # ignore edge case of D==0 for now
        sqrt_D_upper = np.sqrt(D_upper.where(upper_exists, other=np.nan))
        s_upper_entry = - F_dot_rayhat - sqrt_D_upper   # smaller s is always entry
        s_upper_exit  = - F_dot_rayhat + sqrt_D_upper   # larger s is always exit
        # bottom boundary:
        r_lower = self.R0
        D_lower = F_dot_rayhat_2 - F2 + r_lower**2  # discriminant
        lower_exists = (D_lower > 0)  # ignore edge case of D==0 for now
        sqrt_D_lower = np.sqrt(D_lower.where(D_lower >= 0, other=np.nan))
        s_lower_entry = - F_dot_rayhat - sqrt_D_lower
        # s_lower_exit  = - F_dot_rayhat + sqrt_D_lower  # <-- not actually used anywhere
        # combine to get final (min, max).
        s_min = s_upper_entry   # always start with upper entry... no other way to enter.
        s_max = xr.where(
            lower_exists,
            s_lower_entry,   # if lower boundary exists, ray stops at lower entry point.
            s_upper_exit,   # else, ray goes all the way to upper exit point.
        )
        return (s_min, s_max)

    _RAY_S_GRID_DEFAULTS = {
        'name': 'ray_s',
        'inclusive': (False, False),  # avoid including edges, which can lead to NaN during interp.
    }

    def standard_ray_s(self, N):
        '''return standard ray_s with N points along each ray.
        Equivalent: self.ray_s_array() when self.ray_s == N.
        Equivalent: xarray_grid(*self.standard_ray_s_limits(), N=N, **self._RAY_S_GRID_DEFAULTS)
            (self._RAY_S_GRID_DEFAULTS is by default equivalent to: name='ray_s', inclusive=(False, False))
        '''
        smin, smax = self.standard_ray_s_limits()
        return xarray_grid(smin, smax, N=N, **self._RAY_S_GRID_DEFAULTS)

    def ray_s_array(self):
        '''self.ray_s as an xr.DataArray including 'ray_s' or 'ray_s_dim' dimension.
        Behavior depends on self.ray_s:
            xr.DataArray --> if has 'ray_s' or 'ray_s_dim' dim, return as-is, else crash (DimensionalityError)
            None --> crash (InputMissingError)
            int --> use this many points along each ray, with s_min, s_max = self.standard_ray_s_limits()
            dict --> pass kwargs to pc.xarray_grid(), with min, max = self.standard_ray_s_limits() unless provided.
        '''
        result = self.ray_s
        if result is None:
            raise InputMissingError("Cannot get ray_s_array when ray_s is unspecified (self.ray_s=None).")
        elif is_integer(result):
            return self.standard_ray_s(N=result)
        elif isinstance(result, dict):
            if 'min' in result and 'max' in result:
                kw = result
            else:
                smin, smax = self.standard_ray_s_limits()
                kw = {'min': smin, 'max': smax, **self._RAY_S_GRID_DEFAULTS, **result}
            result = xarray_grid(**kw)
        elif not hasattr(result, 'dims'):  # not None, int, dict, nor xarray. Not sure what it is.
            raise InputError("ray_s must be an xarray object including 'ray_s' or 'ray_s_dim' dimension.")
        elif 'ray_s' not in result.dims and 'ray_s_dim' not in result.dims:
            raise DimensionalityError(f"ray_s must have 'ray_s' or 'ray_s_dim' dim, got dims {result.dims}")
        return result

    @property
    def ray_s_dim(self):
        '''return name of ray_s dimension, i.e. 'ray_s' or 'ray_s_dim', depending on which one self.ray_s_array() has.
        [EFF] for now this just checks dims of self.ray_s_array() (i.e. it recomputes that array every time),
            rather than guessing properly before the array is computed.
        '''
        ray_s_array = self.ray_s_array()
        if 'ray_s' in ray_s_array.dims:
            return 'ray_s'
        elif 'ray_s_dim' in ray_s_array.dims:
            return 'ray_s_dim'
        else:
            raise DimensionalityError(f"ray_s_array must have 'ray_s' or 'ray_s_dim' dim, got dims {ray_s_array.dims}")


    # # # PLOTTING # # #

    PLOT_DEFAULTS = {
        ## values related to self.plot() ##
        'box': {
            'bottom': {
                'color': 'orangered',
                'alpha': 0.5,
            },
            'top': {
                'color': 'gray',
                'alpha': 0.5,
            },
            'posts': {
                'color': 'black',
            },
            'A': {
                'color': 'red',
                'marker': r'$\text{A}$',
            },
            'o': {   # marker at xbox=X0, ybox=Y0, zbox=0
                'color': 'orangered',
                'marker': 'o',
                'facecolors': 'none',
            },
            'x': {   # marker at xbox=XEXT, ybox=Y0, zbox=0
                'color': 'orangered',
                'marker': 'x',
            },
        },
        'rays_cmap': 'viridis',
        'rays': {
            'colors': plt.get_cmap('tab10').colors,
        },
        'fov': {
            'color': 'black',
        },
        'view_init': {
            'roll': 270,
            'elev': 15,
            'azim':-15,
        },
        'aspect': (1,1,1),

        ## values related to getting box surfaces/edges/points
        'post_points': 2,  # number of points for each post line connecting box corners
        'surface_points': (21, 21),  # number of points in (u, v) grid for top & bottom
    }

    @format_docstring(**_paramdocs_ray_draping)
    def plot(self, system='both', *, box=True, fov=True, rays=True, max_n_rays=100,
             in_box='lower', periodic=True,
             ax=None, axsize=(6,6), aspect=True, view_init=True,
             extra_points=None, extra_lines=None):
        '''visualize self by plotting. Useful for sanity checks!
        Returns matplotlib ax (or tuple of (ax_sph, ax_box) if system='both').

        If using Jupyter, for easier 3D visualization consider running a cell with command:
            %matplotlib notebook 

        (Below, False-ness checked via "is False", to avoid ambiguity with empty dicts)

        system: 'both', 'sph' or 'box'
            which system(s) to visualize.
            'both' --> make a figure with two plots, one for each system.
            'sph' --> 'sph' system, where rays are straight, and box top and bottom are curved.
            'box' --> 'box' system, where rays are curved, and box top and bottom are flat.
        box: bool or dict of dicts
            whether to plot the box boundaries.
            True --> use default box plotting parameters (self.PLOT_DEFAULTS['box']).
            dict of dicts --> customize box plotting parameters. vals are dict or False.
                Use False to indicate "don't plot this part".
                Skip key to indicate "use default kwargs for this part".
                'bottom': kwargs for ax.plot_surface() for bottom of box (zbox=0 plane)
                'top': kwargs for ax.plot_surface() for top of box (zbox=ZEXT plane)
                'posts': kwargs for ax.plot() for vertical posts connecting bottom/top corners
                'A': kwargs for ax.scatter() for anchor point
                'o': kwargs for ax.scatter() for point at xbox=X0, ybox=Y0, zbox=0
                'x': kwargs for ax.scatter() for point at xbox=XEXT, ybox=Y0, zbox=0
        fov: bool or dict
            whether to plot the FOV points.
            True --> use default FOV plotting parameters (self.PLOT_DEFAULTS['fov']).
            dict --> customize FOV plotting parameters. Can provide any kwargs for ax.scatter().
        rays: bool or dict
            whether to plot the rays.
            True --> use default ray plotting parameters (self.PLOT_DEFAULTS['box']).
            dict --> customize ray plotting parameters. Can provide any kwargs for ax.plot().
            Some kwargs are treated specially: 'colors' and 'cmap'
                if 'colors' is provided, cycle these colors if there are more colors than rays,
                else use 'cmap', or self.PLOT_DEFAULTS['rays_cmap'] if 'cmap' not provided.
        max_n_rays: int or None
            maximum number of rays to plot, else crash (if doing fov and/or rays).
            prevents accidentally plotting way too much if doing production-grade RayDraping.
            None --> no limit.
        in_box: {in_box}
            (The check here applies only to plotted rays)
        periodic: {periodic}
            (The wrapping here applies only to plotted rays and FOV points when system=='box')
            (If periodic wrapping leads to weird-looking lines crossing the box,
                consider using rays={{'linestyle':'none', 'marker':'.'}} to just plot points.)
        ax: None or matplotlib Axes3D object
            Axes to plot on. None --> create new figure and axes.
            Incompatible with system='both'.
        axsize: tuple of two ints
            (x,y) size of new figure, per axes (only if making new axes because ax is None).
            figsize = (2x, y) if system='both', else (x,y).
        aspect: bool or 3-tuple of numbers
            whether to set aspect ratio for 3D axes.
            True --> set aspect based on self.PLOT_DEFAULTS['view_init']. (default: (1,1,1))
            3-tuple --> (x,y,z) scaling to use.
            (Matplotlib is annoying about this, so we have to do:
                sizes=[xmax-xmin, ymax-ymin, zmax-zmin] from get_xlim(), get_ylim(), get_zlim(),
                ax.set_box_aspect(sizes[0]*aspect[0], sizes[1]*aspect[1], sizes[2]*aspect[2]))
        view_init: bool or dict
            whether to ax.view_init().
            True --> use self.PLOT_DEFAULTS['view_init'].
            dict --> dict of kwargs for ax.view_init(). (Consider: 'elev', 'azim', 'roll'.)
        '''
        # checks first (if will crash, crash before plotting anything)
        if system not in ['both', 'sph', 'box']:
            raise InputError(f"system must be 'both', 'sph', or 'box', but got {system!r}")
        if ax is not None and system == 'both':
            raise InputConflictError("Cannot provide ax when system='both'.")
        if (max_n_rays is not None) and (fov is not False or rays is not False):
            if self.n_rays > max_n_rays:
                errmsg = (f'Too many rays to plot: {self.n_rays}. Max is max_n_rays={max_n_rays}.'
                          '\nConsider changing max_n_rays, or reducing number of points in FOV.'
                          '\nUse max_n_rays=None to remove this limit entirely.')
                raise TooManyPlottablesError(errmsg)
        # setup axes
        if ax is None:
            if system == 'both':
                fig = plt.figure(figsize=(axsize[0]*2, axsize[1]))
                ax_sph = fig.add_subplot(1, 2, 1, projection='3d')
                ax_box = fig.add_subplot(1, 2, 2, projection='3d')
                kw_plot = dict(box=box, fov=fov, rays=rays, max_n_rays=max_n_rays,
                               in_box=in_box, periodic=periodic,
                               aspect=aspect, view_init=view_init)
                ax_sph = self.plot('sph', ax=ax_sph, **kw_plot)
                ax_box = self.plot('box', ax=ax_box, **kw_plot)
                return (ax_sph, ax_box)
            else:
                fig = plt.figure(figsize=axsize)
                ax = fig.add_subplot(1, 1, 1, projection='3d')
        assert system != 'both'    # if this assertion fails, made a coding error.
        # do plotting
        if box is not False:
            boxdefaults = self.PLOT_DEFAULTS['box']
            boxkeys = ['bottom', 'top', 'posts', 'o', 'x']
            if isinstance(box, dict):
                boxkws = {k: {**boxdefaults[k], **box.get(k, {})} for k in boxkeys}
            else:
                boxkws = boxdefaults
            boxkws = {k: v for k, v in boxkws.items() if v is not False}
            boxv = self._get_box_plot_values(system=system, include=boxkws.keys())
            if 'A' in boxkws:
                ax.scatter(*xarray_take_xyz(boxv['A']), **boxkws['A'])
            if 'o' in boxkws:
                ax.scatter(*xarray_take_xyz(boxv['o']), **boxkws['o'])
            if 'x' in boxkws:
                ax.scatter(*xarray_take_xyz(boxv['x']), **boxkws['x'])
            if 'posts' in boxkws:
                for post_key in ['post00', 'post10', 'post01', 'post11']:
                    ax.plot(*xarray_take_xyz(boxv[post_key]), **boxkws['posts'])
            if 'bottom' in boxkws:
                ax.plot_surface(*xarray_take_xyz(boxv['bottom']), **boxkws['bottom'])
            if 'top' in boxkws:
                ax.plot_surface(*xarray_take_xyz(boxv['top']), **boxkws['top'])
        if fov is not False:
            fovdefaults = self.PLOT_DEFAULTS['fov']
            fovkws = {**fovdefaults, **fov} if isinstance(fov, dict) else fovdefaults
            FOVv = self.FOV if system == 'sph' else self.sph_to_box(self.FOV, periodic=periodic)
            ax.scatter(*xarray_take_xyz(FOVv), **fovkws)
        if rays is not False:
            raydefaults = self.PLOT_DEFAULTS['rays']
            raykws = {**raydefaults, **rays} if isinstance(rays, dict) else {**raydefaults}
            colors = raykws.pop('colors', None)
            cmap = raykws.pop('cmap', self.PLOT_DEFAULTS['rays_cmap'])
            n_rays = self.n_rays
            if colors is not None:
                if len(colors) < n_rays:
                    colors = None  # not enough colors
            if colors is None:
                cmap = plt.get_cmap(cmap)
                colors = cmap(np.linspace(0, 1, n_rays))
            if system == 'sph':
                rayv = self.wsph(in_box=in_box)
            else:
                rayv = self.wbox(in_box=in_box, periodic=periodic)
            raystack = rayv.stack(_pc_ray_i_=list(set(rayv.dims) - {'component', 'ray_s', 'ray_s_dim'}))
            for i in range(n_rays):
                ray = raystack.isel(_pc_ray_i_=i)
                ax.plot(*xarray_take_xyz(ray), color=colors[i], **raykws)
        # view-related stuff
        ax.set(xlabel='x', ylabel='y', zlabel='z', title=f'{system} system')
        if aspect is not False:
            if aspect is True:
                aspect = self.PLOT_DEFAULTS['aspect']
            lims = [getattr(ax, f'get_{x}lim')() for x in ['x', 'y', 'z']]
            sizes = [lim[1] - lim[0] for lim in lims]
            ax.set_box_aspect((sizes[0]*aspect[0], sizes[1]*aspect[1], sizes[2]*aspect[2]))
        if view_init is not False:
            if not isinstance(view_init, dict):
                view_init = self.PLOT_DEFAULTS['view_init']
            ax.view_init(**view_init)
        return ax

    def _get_box_plot_values(self, *, system='box', include=['o', 'x', 'A', 'posts', 'bottom', 'top']):
        '''return dict of values related to box, for self.plot(box=True).
        Results are xr.DataArrays with 'component' dim telling x,y,z coords.

        system: 'box' or 'sph'
            which system to get box values for.
            'box' --> returns values in box system cartesian coords. No conversion necessary.
            'sph' --> convert to sph system cartesian coords centered at r=0, via self.box_to_sph().
        include: list of strs, tells which keys to include. (Use 'posts' to include all posts):
            'o': point at xbox=X0, ybox=Y0, zbox=0. This is the lower-left point in the box.
            'x': point at xbox=X0+XEXT, ybox=Y0, zbox=0
            'A': the anchor point
            'post00': (xbox=X0,ybox=Y0), zbox from 0 to ZEXT, with dim '_pc_post_i_'
            'post10': (xbox=X0+XEXT,ybox=Y0), zbox from 0 to ZEXT, with dim '_pc_post_i_'
            'post01': (xbox=X0,ybox=Y0+YEXT), zbox from 0 to ZEXT, with dim '_pc_post_i_'
            'post11': (xbox=X0+XEXT,ybox=Y0+YEXT), zbox from 0 to ZEXT, with dim '_pc_post_i_'
            'bottom': zbox=0 surface, with dims '_pc_surface_u_', and '_pc_surface_v_'
            'top': zbox=ZEXT surface, with dims '_pc_surface_u_', and '_pc_surface_v_'

        The number of points along the posts and surface dims are determined by
            self.PLOT_DEFAULTS['post_points'] and self.PLOT_DEFAULTS['surface_points'].
        '''
        result = {}
        BASIS = self._get_basis()
        XHAT = BASIS.XHAT
        YHAT = BASIS.YHAT
        ZHAT = BASIS.ZHAT
        X0 = self.X0
        Y0 = self.Y0
        XEXT = self.XEXT
        YEXT = self.YEXT
        X1 = X0 + XEXT
        Y1 = Y0 + YEXT
        # points at (0,0,0) and (XEXT,0,0)
        if 'o' in include:
            result['o'] = X0 * XHAT + Y0 * YHAT
        if 'x' in include:
            result['x'] = X1 * XHAT + Y0 * YHAT
        if 'A' in include:
            result['A'] = self.Abox_x * XHAT + self.Abox_y * YHAT   # + 0 * ZHAT
        # posts
        if 'posts' in include:
            pspanz = np.linspace(0, self.ZEXT, self.PLOT_DEFAULTS['post_points'])
            pspanz = xr1d(pspanz, name='_pc_post_i_')
            xyposts = {'post00': (X0,Y0), 'post10': (X1,Y0),
                       'post01': (X0,Y1), 'post11': (X1,Y1)}
            result.update({k: x_ * XHAT + y_ * YHAT + pspanz * ZHAT for k, (x_, y_) in xyposts.items()})
        # bottom and top surfaces
        if 'bottom' in include or 'top' in include:
            uu = np.linspace(X0, X1, self.PLOT_DEFAULTS['surface_points'][0])
            vv = np.linspace(Y0, Y1, self.PLOT_DEFAULTS['surface_points'][1])
            uu = xr1d(uu, name='_pc_surface_u_')
            vv = xr1d(vv, name='_pc_surface_v_')
            if 'bottom' in include:
                result['bottom'] = uu * XHAT + vv * YHAT + 0 * ZHAT
            if 'top' in include:
                result['top'] = uu * XHAT + vv * YHAT + self.ZEXT * ZHAT
        # convert system if needed
        if system == 'sph':
            for k, v in result.items():
                result[k] = self.box_to_sph(v)
        return result

    def get_post(self, x, y, system='box', *, dim='_pc_post_i_', N=2):
        '''returns this post, i.e. points corresponding to line from (x,y,0) to (x,y,ZEXT) in box system.
        Result is an xarray.DataArray with 'component' dimension, converted to the indicated system.

        x, y: scalars
            coordinates of post in box system.
            In box system, the post is a line from (x,y,0) to (x,y,ZEXT).
        dim: str
            name of dimension for points along post.
        N: int
            number of points along post (including endpoints).
        '''
        BASIS = self._get_basis()
        zz = xr1d(np.linspace(0, self.ZEXT, N), name=dim)
        post = x * BASIS.XHAT + y * BASIS.YHAT + zz * BASIS.ZHAT
        if system == 'sph':
            post = self.box_to_sph(post)
        return post

    def plot_post(self, x, y, ax_sph=None, ax_box=None, **kw_plot):
        '''plot a post (i.e. line from z=0 to z=ZEXT in box system) on the indicated ax(es).

        x, y: scalars
            coordinates of post in box system.
            In box system, the post is a line from (x,y,0) to (x,y,ZEXT).
        ax_sph, ax_box: None or matplotlib axes objects
            plots post in sph system on ax_sph, in box system on ax_box.
            None --> don't plot on that system.
            At least one of these two inputs must be provided, else crash.
        Additional kwargs go to matplotlib ax.plot() calls.

        Example (plots post at x=5, y=7 in both systems):
            draper = RayDraping(...)
            axes = draper.plot(system='both')
            draper.plot_post(5, 7, *axes)
            # equivalent: draper.plot_post(5, 7, ax_sph=axes[0], ax_box=axes[1])
        '''
        if ax_sph is None and ax_box is None:
            raise InputMissingError("Must provide at least one of ax_sph or ax_box.")
        if ax_sph is not None:
            post_sph = self.get_post(x, y, system='sph')
            ax_sph.plot(*xarray_take_xyz(post_sph), **kw_plot)
        if ax_box is not None:
            post_box = self.get_post(x, y, system='box')
            ax_box.plot(*xarray_take_xyz(post_box), **kw_plot)


    # # # DISPLAY # # #

    def _repr_contents(self):
        '''list of contents for self.__repr__'''
        result = super()._repr_contents()
        if is_integer(self.ray_s) or isinstance(self.ray_s, dict):
            result.append(f"ray_s={self.ray_s}")
        elif self.ray_s is not None:
            result.append(f"ray_s=DataArray(sizes={dict(self.ray_s.sizes)})")
        return result



"""
Package Purpose: registry of runs, including metadata, params, and metrics.
See also: https://gitlab.com/Sevans7/plasmacalcs/-/issues/44
"""

from .metadata_haver import MetadataHaver, pc_uuid_here, UuidHaver
from .register import Registry, Registerable
from .registry_tools import (
    registry_schema_sql,
    registry_location_default, registry_location, registry_name_default,
)



"""
File Purpose: RayDrapingLoader

This provides "field-of-view" grid of rays through a box draped across a sphere.
Intended for use with calculators having maindims = ('x', 'y', 'z').

For more details see tools.sci_tools.ray_draping.
"""

import numpy as np

from .patterns import VectorArithmeticLoader, LoiLoader
from .quantity_loader import QuantityLoader

from ..errors import (
    InputMissingError, InputError,
    DimensionalityError, DimensionValueError,
)
from ..tools import (
    alias, simple_property, format_docstring,
    RayDraping,
    xarray_take_xyz,
)

class RayDrapingLoader(QuantityLoader):
    '''ray draping loader, enabling ability to interpolate to ray paths through box,
    defined by field-of-view grid of rays intersecting box region draped across a sphere.

    For more details see help(self.ray_draping_cls)
    '''

    # # # RAY_DRAPER # # #

    ray_draping_cls = RayDraping

    cls_behavior_attrs.register('ray_draping', default=None)
    ray_draping = simple_property('_ray_draping', default=None,
        doc='''None, dict, or RayDraping instance defining RayDraping for ray draping problems.
        Set kwargs here, then get the RayDraping instance via self.ray_draper. E.g.:
            self.ray_draping=dict(...)  # <-- set RayDraping kwargs as desired
            self.ray_draper.rays(system='box')   # <-- ray paths in "box" system

        None --> quantities needing ray_draping will crash.
        dict --> will use RayDraping.from_dict({**self.ray_draping_defaults(), **ray_draping})
                (e.g., XEXT, YEXT, and ZEXT will come from self.maindims_coords() by default.)
        RayDraping --> will use that RayDraping instance directly.''')

    ray_draper = alias('ray_draping_explicit',
        doc='''The RayDraping instance to use. alias to self.ray_draping_explicit.
        Use self.ray_draper to access the RayDraping instance implied by self.ray_draping,
            regardless of whether it is None, dict, or already a RayDraping instance.
        See help(type(self).ray_draping) for more details.''')

    ray_draping_explicit = alias('ray_draping',
        doc='''self.ray_draping, but convert to RayDraping instance if needed.
        (crashes with InputMissingError if self.ray_draping is None.)''')
    @ray_draping_explicit.getter
    def ray_draping_explicit(self):
        rd = self.ray_draping
        if rd is None:
            errmsg = ('RayDrapingLoader cannot determine RayDraping configuration when ray_draping is None.\n'
                      'Consider setting self.ray_draping = dict(...) or RayDraping(...).\n'
                      '(See help(type(self).ray_draping) for details.)')
            raise InputMissingError(errmsg)
        elif isinstance(rd, RayDraping):
            return rd
        elif isinstance(rd, dict):
            kws = {**self.ray_draping_defaults(), **rd}
            return self.ray_draping_cls.from_dict(kws)
        else:
            errmsg = (f"{type(self).__name__}.ray_draping_explicit expects ray_draping to be"
                      f"None, dict, or RayDraping instance; got object with type={type(rd).__name__}.")
            raise InputError(errmsg)

    def ray_draping_defaults(self):
        '''return dict of defaults to use for ray_draping.
        Asserts that set(self.maindims) == {'x', 'y', 'z'},
            and that self.maindims_coords()['z'] >= 0 everywhere.
            (The latter assumption may be lifted eventually,
                but present RayDraping implementation assumes z=0 at bottom.)

        XEXT, YEXT, ZEXT, X0, Y0 will be taken from self.maindims_coords()
        No other keys included here.
        '''
        assert hasattr(self, 'maindims'), "RayDrapingLoader expects to be inherited by maindims haver."
        # ^^ slightly hacky, but including MainDimensionsHaver in RayDrapingLoader bases
        #       explicitly instead led to hard-to-debug inheritance mro() issue with PlasmaCalculator.
        if set(self.maindims) != {'x', 'y', 'z'}:
            errmsg = (f"{type(self).__name__}.ray_draping_defaults() expects maindims=={{'x','y','z'}} "
                      f"but got maindims={self.maindims}.\nRay draping on non-xyz dims is not yet implemented.\n"
                      "Workaround option: explicitly set ray_draping = RayDraping(...)")
            raise DimensionalityError(errmsg)
        coords = self.get_maindims_coords()
        xmin, xmax = coords['x'].min(), coords['x'].max()
        ymin, ymax = coords['y'].min(), coords['y'].max()
        zmin, zmax = coords['z'].min(), coords['z'].max()
        xext = xmax - xmin
        yext = ymax - ymin
        zext = zmax - zmin
        if coords['z'].min() != 0:
            errmsg = (f"{type(self).__name__}.ray_draping_defaults() expects min(z)==0,\n"
                      f"but got z with min(z)={coords['z'].min()}.\n"
                      "Ray draping with min(z) != 0 is not yet implemented. "
                      "Workaround option: explicitly set ray_draping = RayDraping(...)")
            raise DimensionValueError(errmsg)
        return {'XEXT': xext, 'YEXT': yext, 'ZEXT': zext, 'X0': xmin, 'Y0': ymin}


    # # # LOS & LOI OPTIONS # # #

    LOS_OPTIONS = {**VectorArithmeticLoader.LOS_OPTIONS,
        'ray_draping': '''var_los == self("var_los", los=self.ray_draper.ray_tangents()).
            i.e., use line-of-sight along rays implied by self.ray_draping.''',
        '-ray_draping': '''var_los == self("var_los", los=-self.ray_draper.ray_tangents()).
            i.e., antiparallel to line-of-sight along rays implied by self.ray_draping''',
    }

    @known_var
    @format_docstring(_doc_=VectorArithmeticLoader.get_los.__doc__)
    def get_los(self):
        '''{_doc_}'''
        los = self.los
        if los in ('ray_draping', '-ray_draping'):
            draper = self.ray_draper
            ray_tangents = draper.ray_tangents()
            if los == 'ray_draping':
                result = ray_tangents
            else:
                result = -ray_tangents
        else:
            result = super().get_los()
        return result

    LOI_OPTIONS = {**LoiLoader.LOI_OPTIONS,
        'los': '''behavior implied by self.los. Probably no interpolation, integration along los dimension.
            However, one special case is if los=='ray_draping' or '-ray_draping'.
                In that case, this is equivalent to using self.loi = 'ray_draping'.
            Otherwise, no interpolation, and integration only works if los implies a dim (e.g. 'x')
                (line-of-sight can generally be an array).''',
        'ray_draping': '''var_loi = self("var_loi", loi=(None, self.ray_draper.rays().pc.take_xyz(as_dict=True))).
            i.e., interpolate to loi ray paths implied by self.ray_draping.
            Meanwhile, integration is along the ray dimension ('ray_s' or 'ray_s_dim').''',
        # [TODO] does loi=='ray_draping' actually work? Or does it cause crash when self.los isn't also 'ray_draping'?
    }

    def _loi_is_ray_draping(self):
        '''returns whether self.loi implies to do ray draping.
        True if self.loi == 'ray_draping', or if self.loi == 'los' and los == 'ray_draping' or '-ray_draping'.
        '''
        if self.loi == 'los':
            return isinstance(self.los, str) and self.los in ('ray_draping', '-ray_draping')
        else:
            return self.loi == 'ray_draping'

    def _loi_interp_dict(self):
        '''return dict of arrays to interpolate onto, as implied by self.loi.
        result is None if self.loi implies "no interpolation".
        Crash if invalid self.loi.
        '''
        if self._loi_is_ray_draping():
            result = xarray_take_xyz(self.ray_draper.rays('box'), as_dict=True)
        else:
            result = super()._loi_interp_dict()
        return result

    def _loi_integration_dim(self, *, signed=False):
        '''return dimension along which to integrate, as implied by self.loi.
        Crash with InputMissingError if self.loi implies no integration.

        signed: bool
            whether to start with '-' to imply "cumsum in reverse order" inside loi_cumsum().
            E.g. '-z' --> array.isel(z=slice(None,None,-1)).cumsum('z') instead of array.cumsum('z').
        '''
        if self._loi_is_ray_draping():
            result = self.ray_draper.ray_s_dim
            neg = self.los.startswith('-')
            if signed and neg:
                result = '-' + result
        else:
            result = super()._loi_integration_dim(signed=signed)
        return result

    @property
    def _loi_ustr_default(self):
        '''default unit string to use for loi-related units, e.g. when calculating dloi.
        here, returns 'length' if self._loi_is_ray_draping(), else result from super().
        '''
        return 'length' if self._loi_is_ray_draping() else super()._loi_ustr_default



"""
File Purpose: misc tests for physics using data from eppic hookup in PlasmaCalcs.
These tests require some simulation output.
"""
import os

# setting mp start method to fork (instead of 'spawn') prevents leaked semaphore objects. Not sure why.
# it also makes the tests a bit faster, maybe...
import multiprocessing as mp
mp.set_start_method('fork', force=True)

import numpy as np
import xarray as xr

import PlasmaCalcs as pc
pc.DEFAULTS.pic_ambiguous_unit = dict(u_t=1)  # must be defined before loading an EppicCalculator.

HERE = os.path.dirname(__file__)


### --------------------- can create eppic calculator --------------------- ###

def get_eppic_calculator(**kw_init):
    '''make an eppic calculator'''
    with pc.InDir(os.path.join(HERE, 'test_eppic_tinybox')):
        ec = pc.EppicCalculator.from_here(**kw_init)
    return ec


### --------------------- drifts --------------------- ###

def test_nusn_from_drift():
    '''ensure nusn_from_drift results are reasonable.'''
    ec = get_eppic_calculator()
    margin = 0.25  # 25% margin of error -- this is noisy stuff since box is so small.
    ec.snap = slice(3, None)  # ignore first few snaps - not enough time to reach equil.
    ec.fluid = pc.IONS
    arr = ec('meant_nusn_from_means_hall/nusn')
    assert np.all(1 - margin < arr) and np.all(arr < 1 + margin)
    arr = ec('meant_nusn_from_means_momE/nusn')
    # pederson & ExB drifts contain enough info to get reasonable value for electron too
    ec.fluid = None
    assert np.all(1 - margin < arr) and np.all(arr < 1 + margin)
    arr = ec('meant_nusn_from_means_pederson/nusn')
    assert np.all(1 - margin < arr) and np.all(arr < 1 + margin)
    arr = ec('meant_nusn_from_means_momExB/nusn')
    assert np.all(1 - margin < arr) and np.all(arr < 1 + margin)

def test_get_drift():
    '''ensure drift velocity predictions are close to actual velocities.
    Not exact, due to noise.
    '''
    ec = get_eppic_calculator()
    margin = 0.35  # 35% margin of error -- this is noisy stuff since box is so small.
    ec.snap = slice(3, None)  # ignore first few snaps - not enough time to reach equil.
    arr = ec('mean_mod_u/mean_mod_u_drift')
    assert 1 - margin < arr.min()
    assert arr.max() < 1 + margin



"""
File Purpose: test eppic PlotterManager.
NOTE: plot color scheme would look nicer if you first do: import colorcet
However, tests should still pass even if colorcet import fails.
"""
import os

import PlasmaCalcs as pc
pc.DEFAULTS.pic_ambiguous_unit = dict(u_t=1)  # must be defined before loading an EppicCalculator.

HERE = os.path.dirname(__file__)
TESTS_DIR = os.path.dirname(HERE)
ARTIFACTS = os.path.join(TESTS_DIR, '_test_artifacts', 'eppic_plotter_manager')
if os.path.exists(ARTIFACTS):  # remove any xarray_io artifacts before running tests.
    import shutil; shutil.rmtree(ARTIFACTS)
# <-- don't makedirs here; xarray_save should makedirs as needed.


def get_eppic_calculator(**kw_init):
    '''make an eppic calculator'''
    with pc.InDir(os.path.join(HERE, 'test_eppic_tinybox')):
        ec = pc.EppicCalculator.from_here(**kw_init)
    return ec

### --------------------- sam_plots --------------------- ###

# takes ~15 to 25 seconds
def test_eppic_sam_plots_lowres():
    '''ensures EppicCalculator.sam_plots_lowres doesn't crash.

    This tests a good chunk of PlotterManager functionality,
        by running the initial group of plotters sam uses for quick eppic analysis.

    [TODO] It would also be nice to add some sort of test for correctness too...
    '''
    ec = get_eppic_calculator()
    # for tinybox, the default fft_keep of 0.1 is too small to show anything!
    ec.sam_defaults['fft_calcs']['fft_keep'] = 0.8
    # mini-test: ensure that^ only changed the default for this instance, not all instances.
    assert pc.EppicCalculator.SAM_DEFAULTS['fft_calcs']['fft_keep'] < 0.8
    # we can reduce the dpi slightly (default 100 here); people usually won't look at these plots...
    dpi = 70
    # let's put the plots in ARTIFACTS/sam_plots_lowres
    # '{savename}' is recognized by the ec.sam_plots_lowres
    dst = os.path.join(ARTIFACTS, 'sam_plots_lowres', '{savename}')
    # here, just ensuring it doesn't crash!
    ec.sam_plots_lowres(dst=dst, dpi=dpi)
